!
! Module: lccorr_corrector
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains variables shared by more than 1 module.
!
! NOTE! The present version is part of lccorr_pcms and is ***NOT*** compatible with lccorr.
! 
! $Id: lccorr_corrector.f90,v 1.2 2000/08/11 13:08:45 pvideler Exp $
!- 
module lccorr_corrector

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use cal	! for caltypes -> PsfBinSizeT
  use errorHandling
  use dss
  use dataFilter     	! in selectlib.

  use lccorr_include
  use exposure_map_pcms
  use exposure_aux

  use array_utils	! for trapezoidSum
  use constants
  use coordinate	! for raDecToInst, instToRaw, getPsfImagePixelCorners
  use dal_aux		! for readColDataToFixed
  use dss_aux
  use dss_ranges_aux
!  use exposure
!use l_exposure
  use geometric_types	! for Point2dT and Point2dDbleT
  use intervals_aux	! for orIntervals
  use regridders	! for regridCartesian
  use ssc_misc		! for hexStrToIntBitMask

use dump
!use plot_utils

  implicit none 

  private
 
  public calcCcdWeightsFromTheory
  public calcCcdWeightsFromEvlist
  public calcExposureCorrections
  public recordRateNullsAsGtis

contains 

  !---------------------------------------------------------------------
  subroutine calcCcdWeightsFromTheory(inSetName, timeAttWcs, expInfo&
  , requiredCorrections, ccdWeights, totalBkgWeight, srcRaDeg, srcDecDeg&
  , bkgMapSetName)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set to the correct instrument and filter before
  ! this routine is called.
  !---------------------------------------------------------------------
  !
  ! Global variables accessed (apart from command-line parameter variables):
  !	lowEnergyPsfEnergyFrac
  !	instrumentId
  !	pleaseCorrectDither
  !	numMicropixelDivs
  !	chipMinRawX
  !	chipMaxRawX
  !	chipMinRawY
  !	chipMaxRawY
  !
  !---------------------------------------------------------------------
  ! NOTE! That 'wcs' as fed to calcCcdExpMap below should be the wcs
  ! of the event list X and Y columns. This is needed to specify the
  ! X-Y coordinate system so that a mask from region filtering on these
  ! columns can be constructed.
  !---------------------------------------------------------------------

    implicit none

    character(*),       intent(in)           :: inSetName
    type(localTAWtype), intent(in)           :: timeAttWcs
    type(expInfoType),  intent(in)           :: expInfo(:,:)
    integer(int32),     intent(in)           :: requiredCorrections
    real(single),       intent(out)          :: ccdWeights(&
                                                  size(expInfo,1),&
                                                  size(expInfo,2)),&
                                                totalBkgWeight
    real(double),       intent(in), optional :: srcRaDeg, srcDecDeg
    character(*),       intent(in), optional :: bkgMapSetName

    integer(int8) :: numCcds, numNodes, ccdNum, nodeNum, ccdNodeNum
    logical(bool) :: isPointSource
!, allSpatialMapsBlank
    type(DataSetT) :: inSet, eventSet
    type(BlockT) :: rateBlock
!    type(DssT) :: dssPointer
    type(RangeT) :: EeVrange, calEeVRange
    type(RangeT), pointer :: EeVRanges(:)
    real(double) :: EeVsamples(numEsamples),  srcThetaArcsec, srcPhi
    real(single) ::&
      spectrum(numEsamples), filterTrans(numEsamples), effArea(numEsamples),&
      EeVFunctions(numEsamples), vigEfunctionAtSrc(numEsamples),&
      qeArray(numCanonicalPatterns, numEsamples), totalQe(numEsamples),&
      eNorm, flatNorm, ccdFlatExp, ccdBkgWeight
    real(single), pointer ::&
      psfSamplesCubeTelcoord(:,:,:),&
      psfSamplesCubeCcd(:,:,:),&
      bkgSamplesCubeCcd(:,:,:)
    real(double), pointer :: psfPixelCnrsThetaArcsec(:,:), psfPixelCnrsPhi(:,:)
    type(PsfBinSizeT) :: psfPixelSizeMm
    real(single), allocatable :: ccdExpMap(:,:)
    character(2) :: ccdStr
    integer(int16) :: winCtrRawXi, winCtrRawYi, eI
!character(20) :: dummyStr
    integer :: foreignStatus

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, '  aaa calcCcdWeightsFromTheory'
    call message(AppMsg, Noisy, 'Entering subroutine &
      &calcCcdWeights.')

    numCcds  = size(expInfo,1)
    numNodes = size(expInfo,2)

    ! Construct outSet column names:

    isPointSource = .false. ! default
    if (present(srcRaDeg) .and. present(srcDecDeg)) then
      isPointSource = .true.
if (testFlag) print *, '  srcRaDeg', srcRaDeg
if (testFlag) print *, '  srcDecDeg', srcDecDeg
    end if

    inSet = dataSet(inSetName, READ)
    rateBlock = block(inSet, 'RATE')
!    dssPointer = dataSubspace(rateBlock)

    !...................................................................
    ! Get the flux and response functions (including energy bounds and
    ! sample intervals):

if (testFlag) print *, '  bbb'

!!!    if (!** with spectrum) then
!!!      call getEnergyBounds(dssPointer, rateBlock, EeVrange, spectrum)
!!!    else
!      call getEnergyBounds_1(dssPointer, rateBlock, expInfo(1,1)%temperature&
!        , timeAttWcs%timeStamp, EeVrange) ! in exposure_map.

!EeVrange = getEnergyBounds(dssPointer, rateBlock, expInfo(1,1)%temperature&
!  , timeAttWcs%timeStamp)
! in exposure_map.

      calEeVRange = getEnergyBounds() ! just returns the cal range
      call andRanges(userEeVRanges, calEeVRange, EeVRanges)
      EeVRange%lower = EeVRanges(1)%lower
      EeVRange%upper = EeVRanges(size(EeVRanges))%upper

!!!    end if

if (testFlag) print *, 'EeVrange:', EeVrange%lower%value, 'to', EeVrange%upper%value
if (testFlag) print *, 'EeVrange bound types:', EeVrange%lower%type, EeVrange%upper%type

    call sampleEnergyRange(EeVrange, EeVsamples) ! in ssclib/exposure.
if (testFlag) print *, '  ccc'

!!!    if (!** with spectrum) then
!*** interpolate user spectrum
!!!    else
      call constructModelSpectrum(EeVsamples, spectrum) ! in exposure_map.
!!!    end if
if (testFlag) print *, '  ddd'

    call getFilterTransmission(EeVsamples, filterTrans) ! in ssclib/exposure.
if (testFlag) print *, '  eee'
    call getEffectiveArea(EeVsamples, effArea) ! in ssclib/exposure.
if (testFlag) print *, '  fff'

    EeVFunctions = spectrum * effArea * filterTrans
if (testFlag) print *, '  ggg'

    if (isPointSource) then
      call raDecToTel(srcRaDeg*PiOn180, srcDecDeg*PiOn180, timeAttWcs%attitude&
       , timeAttWcs%timeStamp, srcThetaArcsec, srcPhi)
      ! in ssclib/coordinate.f90

      call calcVigAtSourcePos(srcThetaArcsec, srcPhi, EeVsamples&
        , vigEfunctionAtSrc) ! in ssclib/exposure.f90

      call CAL_SetState(accuracyLevel=ACCURACY_MEDIUM)

      call constructPsfCubeTelcoord(EeVsamples, srcThetaArcsec, srcPhi&
        , lowEnergyPsfEnergyFrac, psfSamplesCubeTelcoord, psfPixelSizeMm)
      ! in ssclib/exposure.

if (testFlag) print *, '  hhh'
if (testFlag) call dumpImageToFits(psfSamplesCubeTelcoord(1,:,:), 'psf.ds')

      allocate(psfPixelCnrsThetaArcsec(&
        size(psfSamplesCubeTelcoord,2)+1,&
        size(psfSamplesCubeTelcoord,3)+1)&
      )
      allocate(psfPixelCnrsPhi(&
        size(psfSamplesCubeTelcoord,2)+1,&
        size(psfSamplesCubeTelcoord,3)+1)&
      )
      call getPsfPixelCnrsThetaPhi(srcThetaArcsec, srcPhi&
        , psfPixelSizeMm, psfSamplesCubeTelcoord(1,:,:)&
        , psfPixelCnrsThetaArcsec, psfPixelCnrsPhi)
      ! in ssclib/coordinate.

!***** write psf array size to output table

!**** check Hermann's claim that the telcoord coords are inverted

      EeVFunctions = EeVFunctions * vigEfunctionAtSrc

    else
      vigEfunctionAtSrc = 1.0
    end if
if (testFlag) print *, '  iii'

    !...................................................................
    ! Other init:

call hexStrToIntBitMask(flagMaskHexStr, flagMask, foreignStatus) ! in ssclib/ssc_misc.
if (foreignStatus .ne. 0) call error('badHexToIntStatus',"There was an error &
&in the conversion of the --flagmask parameter from hexadecimal to integer.")

    allocate(ccdExpMap(chipMinRawX:chipMaxRawX, chipMinRawY:chipMaxRawY))

call message(AppMsg, Verbose, 'Opening the event set.')
eventSet = dataSet(eventSetName, READ)

!    allSpatialMapsBlank = .true. ! default
    totalBkgWeight = 0.0
    ccdWeights = 0.0

if (testFlag) print *, '  jjj'
    do ccdNum = 1, numCcds
      do nodeNum = 1, numNodes
      if (.not. expInfo(ccdNum, nodeNum)%exposuExists) cycle
!if (testFlag .and. ccdNum .ne. 3) cycle

        if (instrumentId == EPN) then ! just to stop the cal giving a warning:
          call CAL_setState(ccdChipId=int(ccdNum, int32))
        else
          call CAL_setState(ccdChipId=int(ccdNum, int32)&
            , ccdNodeId=int(nodeNum - 1, int32))
        end if
if (testFlag) print *, '  kkk'

        ccdNodeNum = 10 * (nodeNum - 1) + ccdNum
        write(ccdStr, '(i2.2)') ccdNodeNum
        call message(AppMsg, Verbose, "********************************")
        call message(AppMsg, Verbose, "Found EXPOSUnn extension for &
          &CCD/node number nn = "//ccdStr)
if (testFlag) print *, '  lll'

        winCtrRawXi = nint((expInfo(ccdNum,nodeNum)%winMinRawX&
          + expInfo(ccdNum,nodeNum)%winMaxRawX) / 2.0)
        winCtrRawYi = nint((expInfo(ccdNum,nodeNum)%winMinRawY&
          + expInfo(ccdNum,nodeNum)%winMaxRawY) / 2.0)
if (testFlag) print *, '  mmm'

        call constructQeArray(EeVsamples, qeArray, winCtrRawXi, winCtrRawYi)
        ! in ssclib/exposure.f90
if (testFlag) print *, '  nnn'

        eNorm = calcEeVFunctionNorm(EeVsamples, spectrum * effArea, filterTrans&
          , vigEfunctionAtSrc, qeArray, requiredCorrections) ! in exposure_map
if (testFlag) print *, '  ooo'

        totalQe = (/(sum(qeArray(:,eI)), eI=1,numESamples)/)
        flatNorm = integrateRanges(real(EeVSamples, single)&
          , EeVfunctions * totalQe, (/EeVrange/))

        !...............................................................
        ! The first step is to call calcCcdExpMap. This subroutine takes a
        ! pointer to the data subspace, a spectrum and (optionally) a flux
        ! cube, and returns both an exposure map and a 'flat exposure'
        ! value. The latter is the result of an internal integration in
        ! the spatial coordinates of a second exposure map formed under
        ! the assumption that the event flux is spatially uniform. This
        ! can be useful for estimating background exposure if the background
        ! map is not known.
        !
        ! The aim of this part is to calculate two numbers: the actual
        ! exposure weight for this CCD, and the weight that would have
        ! resulted in the case of a background event flux. (If the time
        ! series is the background one then these two numbers will have
        ! the same value; if the TS is from the source itself, then they
        ! will be different.) The exposure weight is needed for exposure
        ! correction; the background exposure weight will be needed
        ! externally to scale the background TS to the same collection
        ! area as the source TS.

        if (isPointSource) then
if (testFlag) print *, '  ppp'
          call interpolatePsfCubeToCcd(psfPixelCnrsPhi, psfPixelCnrsThetaArcsec&
            , psfSamplesCubeTelcoord, psfSamplesCubeCcd)!, testFlagArg=ptf)

          ! Note to later developers: DON'T be tempted to dump out of the
          ! loop at this point if no part of the psf overlaps the chip.
          ! We still need to see if any of the spatial selection region
          ! overlaps the chip in order to calculate the BKGRATIO value.
if (testFlag) print *, '  qqq'

!          call calcCcdExpMap(eventSet, ccdNodeNum, dssPointer&
!            , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
!            , expInfo(ccdNum,nodeNum), EeVsamples, EeVFunctions, qeArray&
!            , psfSamplesCubeCcd, ccdExpMap, ccdFlatExp&
!          )
          call calcCcdExpMap(tempEventListSetName, eventSet, ccdNodeNum&
            , rawYLo, patternLo, patternHi, flagMask, EeVRanges&
            , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
            , expInfo(ccdNum, nodeNum), EeVsamples, EeVfunctions, qeArray&
            , psfSamplesCubeCcd, ccdExpMap, ccdFlatExp&
            , regionTabName=srcRegionTabName&
          )
          ! in exposure_map_pcms.
!print *, 'just got ccdExpMap'
!write(dummyStr, '(a,i2.2,a)') 'expmap_ccd', ccdNodeNum, '.fits'
!call dumpImageToFits(ccdExpMap, dummyStr)

          deallocate(psfSamplesCubeCcd)
          ccdWeights(ccdNum,nodeNum) = sum(ccdExpMap) / eNorm

          if (present(bkgMapSetName)) then
!*** call getBkgCube(bkgMapSetName, ccdNodeNum, bkgSamplesCubeCcd)
!*** also supply instrumentId
call error('',"")

!            call calcCcdExpMap(eventSet, ccdNodeNum, dssPointer&
!              , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
!              , expInfo(ccdNum,nodeNum), EeVsamples, EeVFunctions, qeArray&
!              , bkgSamplesCubeCcd, ccdExpMap, ccdFlatExp&
!            )
            call calcCcdExpMap(tempEventListSetName, eventSet, ccdNodeNum&
              , rawYLo, patternLo, patternHi, flagMask, EeVRanges&
              , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
              , expInfo(ccdNum, nodeNum), EeVsamples, EeVfunctions, qeArray&
              , bkgSamplesCubeCcd, ccdExpMap, ccdFlatExp&
!**** src or bkg?              , regionTabName=bkgRegionTabName&
            )
            ! in exposure_map_pcms.

            deallocate(bkgSamplesCubeCcd)
            ccdBkgWeight = sum(ccdExpMap) / eNorm !*** / flatNorm instead??
          else
            ccdBkgWeight = ccdFlatExp / flatNorm
          end if

        else ! this is the bkg TS not the source TS.
          if (present(bkgMapSetName)) then
!*** call getBkgCube(bkgMapSetName, ccdNodeNum, bkgSamplesCubeCcd)
!*** also supply instrumentId
call error('',"")

!            call calcCcdExpMap(eventSet, ccdNodeNum, dssPointer&
!              , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
!              , expInfo(ccdNum,nodeNum), EeVsamples, EeVFunctions, qeArray&
!              , bkgSamplesCubeCcd, ccdExpMap, ccdFlatExp&
!            )
            call calcCcdExpMap(tempEventListSetName, eventSet, ccdNodeNum&
              , rawYLo, patternLo, patternHi, flagMask, EeVRanges&
              , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
              , expInfo(ccdNum, nodeNum), EeVsamples, EeVfunctions, qeArray&
              , bkgSamplesCubeCcd, ccdExpMap, ccdFlatExp&
!**** src or bkg?              , regionTabName=bkgRegionTabName&
            )
            ! in exposure_map_pcms.

            deallocate(bkgSamplesCubeCcd)
            ccdWeights(ccdNum,nodeNum) = sum(ccdExpMap) / eNorm
            ccdBkgWeight = ccdWeights(ccdNum,nodeNum)!*** / flatNorm instead??

          else
!            call calcCcdExpMap(eventSet, ccdNodeNum, dssPointer&
!              , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
!              , expInfo(ccdNum,nodeNum), EeVsamples, EeVFunctions, qeArray&
!              , ccdExpMap=ccdExpMap, ccdFlatExp=ccdFlatExp&
!            )
            call calcCcdExpMap(tempEventListSetName, eventSet, ccdNodeNum&
              , rawYLo, patternLo, patternHi, flagMask, EeVRanges&
              , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
              , expInfo(ccdNum, nodeNum), EeVsamples, EeVfunctions, qeArray&
              , ccdExpMap=ccdExpMap, ccdFlatExp=ccdFlatExp&
              , regionTabName=bkgRegionTabName&
            )
            ! in exposure_map_pcms.

            ccdWeights(ccdNum,nodeNum) = ccdFlatExp / eNorm
            ccdBkgWeight = ccdFlatExp / flatNorm
          end if
        end if
        totalBkgWeight = totalBkgWeight + ccdBkgWeight
if (testFlag) print *, 'ccdFlatExp, flatNorm:', ccdFlatExp, flatNorm
if (testFlag) print *, 'ccdBkgWeight, total bkgWeight:', ccdBkgWeight, totalBkgWeight

      end do ! loop over nodes
    end do ! loop over ccds
if (testFlag) print *, '  rrr'

call release(eventSet)

    call release(inSet)
!    call dssRelease(dssPointer)

    call message(AppMsg, Verbose, "********************************")

    if (isPointSource) then
      deallocate(psfSamplesCubeTelcoord)
      deallocate(psfPixelCnrsThetaArcsec, psfPixelCnrsPhi)
    end if

    deallocate(EeVRanges)

if (testFlag) print *, '  zzz calcCcdWeightsFromTheory'

  end subroutine calcCcdWeightsFromTheory

  !---------------------------------------------------------------------
  subroutine calcCcdWeightsFromEvlist(inSetName, timeAttWcs, expInfo&
!  , requiredCorrections&
  , ccdWeights, totalBkgWeight, srcRaDeg, srcDecDeg&
  , bkgMapSetName)
  !---------------------------------------------------------------------
  ! NOTE! The cal must be set to the correct instrument and filter before
  ! this routine is called.
  !---------------------------------------------------------------------
  !
  ! Global variables accessed (apart from command-line parameter variables):
  !	instrumentId
  !	pleaseCorrectDither
  !	numMicropixelDivs
  !	chipMinRawX
  !	chipMaxRawX
  !	chipMinRawY
  !	chipMaxRawY
  !
  !---------------------------------------------------------------------
  ! NOTE! That 'wcs' as fed to calcCcdExpMap below should be the wcs
  ! of the event list X and Y columns. This is needed to specify the
  ! X-Y coordinate system so that a mask from region filtering on these
  ! columns can be constructed.
  !---------------------------------------------------------------------

    implicit none

    character(*),       intent(in)           :: inSetName
    type(localTAWtype), intent(in)           :: timeAttWcs
    type(expInfoType),  intent(in)           :: expInfo(:,:)
!    integer(int32),     intent(in)           :: requiredCorrections
    real(single),       intent(out)          :: ccdWeights(&
                                                  size(expInfo,1),&
                                                  size(expInfo,2)),&
                                                totalBkgWeight
    real(double),       intent(in)           :: srcRaDeg, srcDecDeg
    character(*),       intent(in), optional :: bkgMapSetName

    integer(int8) :: numCcds, numNodes, ccdNum, nodeNum, ccdNodeNum
!    logical(bool) :: isPointSource
    type(DataSetT) :: inSet, eventSet
    type(BlockT) :: rateBlock
    type(RangeT) :: EeVrange, calEeVRange
    type(RangeT), pointer :: EeVRanges(:)
    real(double) :: EeVsamples(numEsamples),  srcThetaArcsec, srcPhi
    real(single) ::&
      spectrum(numEsamples), filterTrans(numEsamples), effArea(numEsamples),&
      EeVFunctions(numEsamples), vigEfunctionAtSrc(numEsamples),&
      qeArray(numCanonicalPatterns, numEsamples), totalQe(numEsamples),&
!      eNorm,&
      flatNorm, ccdFlatExp, ccdBkgWeight
!    real(single), pointer ::&
!      psfSamplesCubeTelcoord(:,:,:),&
!      psfSamplesCubeCcd(:,:,:),&
!      bkgSamplesCubeCcd(:,:,:)
!    real(double), pointer :: psfPixelCnrsThetaArcsec(:,:), psfPixelCnrsPhi(:,:)
!    type(PsfBinSizeT) :: psfPixelSizeMm
    real(single), allocatable :: ccdExpMap(:,:)
    character(2) :: ccdStr
    integer(int16) :: winCtrRawXi, winCtrRawYi, eI
    type(TableT) :: eventTab
    character(256) :: selExpression
    integer :: numRowsSel, foreignStatus
    type(ColumnT) :: ccdCol
    integer(int8), pointer :: ccdData(:)
!character(20) :: dummyStr

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, '  aaa calcCcdWeightsFromEvlist'
    call message(AppMsg, Noisy, 'Entering subroutine &
      &calcCcdWeights.')

    numCcds  = size(expInfo,1)
    numNodes = size(expInfo,2)

    ! Construct outSet column names:

if (testFlag) print *, '  srcRaDeg', srcRaDeg
if (testFlag) print *, '  srcDecDeg', srcDecDeg

    inSet = dataSet(inSetName, READ)
    rateBlock = block(inSet, 'RATE')

    !...................................................................
    ! Get the flux and response functions (including energy bounds and
    ! sample intervals):

if (testFlag) print *, '  bbb'

    calEeVRange = getEnergyBounds() ! just returns the cal range
    call andRanges(userEeVRanges, calEeVRange, EeVRanges)
    EeVRange%lower = EeVRanges(1)%lower
    EeVRange%upper = EeVRanges(size(EeVRanges))%upper

if (testFlag) print *, 'EeVrange:', EeVrange%lower%value, 'to', EeVrange%upper%value
if (testFlag) print *, 'EeVrange bound types:', EeVrange%lower%type, EeVrange%upper%type

    call sampleEnergyRange(EeVrange, EeVsamples) ! in ssclib/exposure.
if (testFlag) print *, '  ccc'

    call constructModelSpectrum(EeVsamples, spectrum) ! in exposure_map.
if (testFlag) print *, '  ddd'

    call getFilterTransmission(EeVsamples, filterTrans) ! in ssclib/exposure.
if (testFlag) print *, '  eee'
    call getEffectiveArea(EeVsamples, effArea) ! in ssclib/exposure.
if (testFlag) print *, '  fff'

    EeVFunctions = spectrum * effArea * filterTrans
if (testFlag) print *, '  ggg'

    call raDecToTel(srcRaDeg*PiOn180, srcDecDeg*PiOn180, timeAttWcs%attitude&
     , timeAttWcs%timeStamp, srcThetaArcsec, srcPhi)
    ! in ssclib/coordinate.f90

    call calcVigAtSourcePos(srcThetaArcsec, srcPhi, EeVsamples&
      , vigEfunctionAtSrc) ! in ssclib/exposure.f90

    EeVFunctions = EeVFunctions * vigEfunctionAtSrc

if (testFlag) print *, '  iii'

    !...................................................................
    ! Other init:

    allocate(ccdExpMap(chipMinRawX:chipMaxRawX, chipMinRawY:chipMaxRawY))

    call message(AppMsg, Verbose, 'Opening the event set.')
!    eventSet = dataSet(eventSetName, READ)
eventSet = clone(eventSetName, tempEventListSetName, TEMP)
    totalBkgWeight = 0.0
    ccdWeights = 0.0

call hexStrToIntBitMask(flagMaskHexStr, flagMask, foreignStatus) ! in ssclib/ssc_misc.
if (foreignStatus .ne. 0) then
print *, 'status=', foreignStatus
call error('badHexToIntStatus',"There was an error &
&in the conversion of the --flagmask parameter from hexadecimal to integer.")
end if
    !...................................................................
    ! Calculate the CCD weights by summing events on each ccd:

    eventTab = table(eventSet, 'EVENTS')
    selExpression = 'region('//trim(srcRegionTabName)//',X,Y)'
    call initFilter(selExpression) 
    numRowsSel = doFilter(eventTab)

    ccdCol = column(eventTab, 'CCDNR')
    ccdData => int8Data(ccdCol)
    do ccdNum = 1, numCcds
      do nodeNum = 1, numNodes
      if (.not. expInfo(ccdNum, nodeNum)%exposuExists) cycle

        ccdNodeNum = 10 * (nodeNum - 1) + ccdNum
        ccdWeights(ccdNum,nodeNum) = count(ccdData==ccdNodeNum)
      end do
    end do
    call release(ccdCol)
    call release(eventTab)
    call release(eventSet)

    !...................................................................
    ! Calculate background area (and normalize ccd weights):

    eventSet = dataSet(eventSetName, READ)

if (testFlag) print *, '  jjj'
    do ccdNum = 1, numCcds
      do nodeNum = 1, numNodes
      if (.not. expInfo(ccdNum, nodeNum)%exposuExists) cycle

        if (instrumentId == EPN) then ! just to stop the cal giving a warning:
          call CAL_setState(ccdChipId=int(ccdNum, int32))
        else
          call CAL_setState(ccdChipId=int(ccdNum, int32)&
            , ccdNodeId=int(nodeNum - 1, int32))
        end if
if (testFlag) print *, '  kkk'

        ccdNodeNum = 10 * (nodeNum - 1) + ccdNum
        write(ccdStr, '(i2.2)') ccdNodeNum
        call message(AppMsg, Verbose, "********************************")
        call message(AppMsg, Verbose, "Found EXPOSUnn extension for &
          &CCD/node number nn = "//ccdStr)
if (testFlag) print *, '  lll'

        winCtrRawXi = nint((expInfo(ccdNum,nodeNum)%winMinRawX&
          + expInfo(ccdNum,nodeNum)%winMaxRawX) / 2.0)
        winCtrRawYi = nint((expInfo(ccdNum,nodeNum)%winMinRawY&
          + expInfo(ccdNum,nodeNum)%winMaxRawY) / 2.0)
if (testFlag) print *, '  mmm'

        call constructQeArray(EeVsamples, qeArray, winCtrRawXi, winCtrRawYi)
        ! in ssclib/exposure.f90
if (testFlag) print *, '  nnn'

!        eNorm = calcEeVFunctionNorm(EeVsamples, spectrum * effArea, filterTrans&
!          , vigEfunctionAtSrc, qeArray, requiredCorrections) ! in exposure_map
if (testFlag) print *, '  ooo'

        totalQe = (/(sum(qeArray(:,eI)), eI=1,numESamples)/)
        flatNorm = integrateRanges(real(EeVSamples, single)&
          , EeVfunctions * totalQe, (/EeVrange/))

        !...............................................................
        ! The first step is to call calcCcdExpMap. This subroutine takes a
        ! pointer to the data subspace, a spectrum and (optionally) a flux
        ! cube, and returns both an exposure map and a 'flat exposure'
        ! value. The latter is the result of an internal integration in
        ! the spatial coordinates of a second exposure map formed under
        ! the assumption that the event flux is spatially uniform. This
        ! can be useful for estimating background exposure if the background
        ! map is not known.
        !
        ! The aim of this part is to calculate two numbers: the actual
        ! exposure weight for this CCD, and the weight that would have
        ! resulted in the case of a background event flux. (If the time
        ! series is the background one then these two numbers will have
        ! the same value; if the TS is from the source itself, then they
        ! will be different.) The exposure weight is needed for exposure
        ! correction; the background exposure weight will be needed
        ! externally to scale the background TS to the same collection
        ! area as the source TS.

if (testFlag) print *, '  ppp'
if (testFlag) print *, '  qqq'

        call calcCcdExpMap(tempEventListSetName, eventSet, ccdNodeNum&
          , rawYLo, patternLo, patternHi, flagMask, EeVRanges&
          , timeAttWcs, pleaseCorrectDither, numMicropixelDivs&
          , expInfo(ccdNum, nodeNum), EeVsamples, EeVfunctions, qeArray&
          , ccdExpMap=ccdExpMap, ccdFlatExp=ccdFlatExp&
          , regionTabName=srcRegionTabName&
        ) ! in exposure_map_pcms.

if (testFlag) call error('testHalt',"")

!        ccdWeights(ccdNum,nodeNum) = sum(ccdExpMap) / eNorm

        if (present(bkgMapSetName)) then
call error('notYetSupported',"")
        else
          ccdBkgWeight = ccdFlatExp / flatNorm
        end if

        totalBkgWeight = totalBkgWeight + ccdBkgWeight
if (testFlag) print *, 'ccdFlatExp, flatNorm:', ccdFlatExp, flatNorm
if (testFlag) print *, 'ccdBkgWeight, total bkgWeight:', ccdBkgWeight, totalBkgWeight

      end do ! loop over nodes
    end do ! loop over ccds
if (testFlag) print *, '  rrr'

call release(eventSet)

    call release(inSet)

    call message(AppMsg, Verbose, "********************************")

    deallocate(EeVRanges)

if (testFlag) print *, '  zzz calcCcdWeightsFromEvlist'

  end subroutine calcCcdWeightsFromEvlist

  !---------------------------------------------------------------------
  subroutine calcExposureCorrections(inSetName, outSetName, eventSetName&
  , expInfo, requiredCorrections, isPointSource, applyCorrections, ccdWeights)
  !---------------------------------------------------------------------

    implicit none

    character(*),      intent(in) :: inSetName, outSetName, eventSetName
    type(expInfoType), intent(in) :: expInfo(:,:)
    integer(int32),    intent(in) :: requiredCorrections
    logical,           intent(in) :: isPointSource
    logical,           intent(in) :: applyCorrections
    real(single),      intent(in) :: ccdWeights(&
                                       size(expInfo,1),&
                                       size(expInfo,2))

    integer(int8) :: numCcds, numNodes, ccdNum, nodeNum, ccdNodeNum
    character(8) ::&
      rateColName, errorColName, fracExpColName, outGtiTabName, exposuTabName
    type(DataSetT) :: inSet, outSet, eventSet
    type(BlockT) :: rateBlock
!    type(DssT) :: dssPointer
    integer(int32) :: numRows, row
    real(single), allocatable ::&
      fracExp(:), ccdFracExp(:), ccdAntiAlias(:), antiAlias(:)
    logical(bool), allocatable :: badBins(:)
    logical(bool) :: someEventsNotRandomized
    type(IntervalT), pointer :: gtis(:)
    character(2) :: ccdStr
    type(TableT) :: outTab
    real(single), pointer :: tsFracExp(:), rate(:), uncert(:)
    type(ColumnT) :: rateCol, errorCol
    real(single) :: totalCcdWeight
integer :: i
logical(bool) ::&
  treatGti, treatCosmicRays, treatDeadTime, treatAlias
!**** Neaten this up some time.

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Noisy, 'Entering subroutine &
      &calcExposureCorrections.')

    numCcds  = size(expInfo,1)
    numNodes = size(expInfo,2)

treatGti         = btest(requiredCorrections, TREAT_GTI_BIT)
treatCosmicRays  = btest(requiredCorrections, TREAT_COSMIC_BIT)
treatDeadTime    = btest(requiredCorrections, TREAT_DEAD_TIME_BIT)
treatAlias       = btest(requiredCorrections, TREAT_ALIAS_BIT)
!**** Neaten this up some time. ALSO! Shouldn't use names of globals.

    ! Construct outSet column names:

    if (isPointSource) then
      rateColName    = 'RATE'
      errorColName   = 'ERROR'
      fracExpColName = 'FRACEXP'
      outGtiTabName  = 'SRC_GTIS'
    else
      rateColName    = 'BRATE'
      errorColName   = 'BERROR'
      fracExpColName = 'BFRACEXP'
      outGtiTabName  = 'BKG_GTIS'
    end if
if (testFlag) print *, 'col/ext names: '&
//trim(rateColName)//' '//trim(errorColName)//' '//trim(fracExpColName)&
//' '//trim(outGtiTabName)

    inSet = dataSet(inSetName, READ)
    rateBlock = block(inSet, 'RATE')
!    dssPointer = dataSubspace(rateBlock)

    outSet = dataSet(outSetName, MODIFY)
    outTab = table(outSet, 'RATE')
    numRows = numberOfRows(outTab)
    allocate(fracExp(numRows))
    call readColDataToFixed(outTab, fracExpColName, fracExp)
    call release(outSet)

    fracExp = 0.0
    allocate(ccdFracExp(numRows), ccdAntiAlias(numRows), antiAlias(numRows))
    allocate(badBins(numRows))
    antiAlias = 0.0
    badBins = .false.

    someEventsNotRandomized = .false. ! default
!    allocate(oredGtis(0))

if (testFlag) print *, 'hhh'
    eventSet = dataSet(eventSetName, READ)
    totalCcdWeight = 0.0 ! default
    do ccdNum = 1, numCcds ! numCcds, numNodes are declared in lccorr_include
      do nodeNum = 1, numNodes ! and their values set in miscSetups.
      if (ccdWeights(ccdNum, nodeNum) <= 0.0) cycle

        if (.not. expInfo(ccdNum, nodeNum)%exposuExists) then
          write(errStr, '(a, i2.2, a, i1, a)') "The weight for CCD "&
            , ccdNum, " node ", nodeNum-1, " is non-zero, but there is &
            &no corresponding EXPOSU extension."
          call warning('badWeightList')
      cycle
        end if

        totalCcdWeight = totalCcdWeight + ccdWeights(ccdNum, nodeNum)

        ccdNodeNum = 10 * (nodeNum - 1) + ccdNum
        write(ccdStr, '(i2.2)') ccdNodeNum
        exposuTabName = 'EXPOSU'//ccdStr

        call message(AppMsg, Verbose, "Calculating time-dependent factors &
          &for CCD/node number nn = "//ccdStr)

        write(errstr, '(a, f6.4)') "Weight for this CCD is "&
          , ccdWeights(ccdNum, nodeNum)
        call message(AppMsg, Verbose)

        ccdFracExp   = 1.0 ! default
        ccdAntiAlias = 1.0 ! default

!        if (hasTimeFilter(dssPointer, ccdNodeNum)) then
!**** check that the time filter is of gti type.
!          call readGtisFromDss(inSet, rateBlock, int(ccdNodeNum), gtis)
        if (withGtis) then
          call readGtis(userGtiSetTabNames(ccdNum, nodeNum), gtis)
!*** in ssclib/intervals_aux since ssclib-3.21.

          call correctForGtiAndAlias(eventSet, exposuTabName, outSetName&
            , gtis, treatGti, treatCosmicRays, treatDeadTime, treatAlias&
            , expInfo(ccdNum, nodeNum)%tFrame, ccdFracExp, ccdAntiAlias&
            , badBins)
          ! in exposure_aux.

          deallocate(gtis)
        end if ! hasTimeFilter

if (testFlag) then
  print *, 'ccdNodeNum', ccdNodeNum
  print *, 'fracExp, ccdFracExp, ccdWeights:', fracExp(1), ccdFracExp(1), ccdWeights(ccdNum, nodeNum)
  print *, 'antiAlias, ccdAntiAlias, ccdWeights:', antiAlias(1), ccdAntiAlias(1), ccdWeights(ccdNum, nodeNum)
  if (any(ccdFracExp > 1.0)) then
    i = smallestTrueIndex(ccdFracExp > 1.0)
    print *, '!!!!! ccdFracExp(',i,')=',ccdFracExp(i)
  end if
end if

        fracExp   = fracExp   + ccdFracExp   * ccdWeights(ccdNum, nodeNum)
        antiAlias = antiAlias + ccdAntiAlias * ccdWeights(ccdNum, nodeNum)

        if (any(ccdAntiAlias < 1.0)) someEventsNotRandomized = .true.
!*** check that this works.
      end do ! loop over nodes
    end do ! loop over ccds
    call release(eventSet)

!    if (.not. isPointSource) then
!      fracExp = fracExp / bkgWeight
!      antiAlias = antiAlias / bkgWeight
!    end if

    call release(inSet)
!    call dssRelease(dssPointer)

    if (totalCcdWeight <= 0.0)&
call error('totalWeightIsZero',"The sum of the CCD weights is zero - &
&cannot perform exposure correction.")

    fracExp   = fracExp   / totalCcdWeight
    antiAlias = antiAlias / totalCcdWeight

    where(badBins) fracExp = 0.0
if (testFlag) print *, 'count(badBins)=', count(badBins)

    if (someEventsNotRandomized) then
      fracExp = fracExp * antiAlias
    end if

    !....................................................................
    ! Apply exposure corrections:

    outSet = dataSet(outSetName, MODIFY)
    outTab = table(outSet, 'RATE')
    tsFracExp => real32Data(column(outTab, fracExpColName))
    tsFracExp = fracExp
if (testFlag) print *, 'count(tsFracExp<=0)=', count(tsFracExp<=0)

    if (applyCorrections) then
      call message(AppMsg, Verbose, 'Applying exposure corrections.')

      rateCol  = column(outTab, rateColName)
      errorCol = column(outTab, errorColName)
      rate  => real32data(rateCol)
      uncert => real32data(errorCol)

      do row = 0, numRows - 1
        if (tsFracExp(row) > 0.0) then
          rate(row)   = rate(row)   / tsFracExp(row)
          uncert(row) = uncert(row) / tsFracExp(row)
        else
if (testFlag) print *, 'set to null row', row
          call setToNull(rateCol,  row)
          call setToNull(errorCol, row)
        end if
      end do ! next row.
    end if ! applyCorrections.

!    if (size(oredGtis) > 0) then
!      outGtiTab = addTable(outSet, outGtiTabName, size(oredGtis))
!      startVals => real64Data(addColumn(outGtiTab, 'START', REAL64, 's'))
!      stopVals  => real64Data(addColumn(outGtiTab, 'STOP',  REAL64, 's'))
!      startVals = oredGtis%lower
!      stopVals  = oredGtis%upper
!    end if

    call release(outSet)

  end subroutine calcExposureCorrections

  !---------------------------------------------------------------------
  subroutine interpolatePsfCubeToCcd(psfPixelCnrsPhi, psfPixelCnrsThetaArcsec&
  , psfSamplesCubeTelcoord, psfSamplesCubeCcd, doDummy)
!, testFlagArg)
  !---------------------------------------------------------------------
  ! NOTE that this routine requires that the cal have been set up by the
  ! calling routine for the correct ccd, node number and instrument.
  !
  ! NOTE psfSamplesCubeCcd should be deallocated after use.

!*** can a coordinate routine be used??

    implicit none 

    real(double),      intent(in)  :: psfPixelCnrsPhi(:,:),&
                                      psfPixelCnrsThetaArcsec(&
                                        size(psfPixelCnrsPhi,1),&
                                        size(psfPixelCnrsPhi,2)&
                                      )
    real(single),      intent(in)  :: psfSamplesCubeTelcoord(:,:,:)

    real(single),      pointer     :: psfSamplesCubeCcd(:,:,:)
    logical, intent(in), optional  :: doDummy
!, testFlagArg

!    type(Point2dT) ::&
!      psfImgPixCrnrRaw(size(psfPixelCnrsPhi,1), size(psfPixelCnrsPhi,2))
type(Point2dDbleT), allocatable :: psfImgPixCrnrRaw(:,:)
type(Point2dT), allocatable :: psfImgPixCrnrSng(:,:)
    ! These data types are defined in ssclib/geometric_types.

    integer(int16) :: minRawXi, maxRawXi, minRawYi, maxRawYi, numEsamples, eI
    integer :: foreignStatus
    logical(bool), allocatable :: isOffChip(:,:)
    real(single) :: norm
!real(double), allocatable :: dummyX(:,:), dummyY(:,:)

logical :: testFlag = .true.
!integer :: loxi, hixi, loyi, hiyi

    !-------------------------------------------------------------------
    ! Start:

!if (present(testFlagArg)) then
!  testFlag = testFlagArg
!else
!  testFlag = .false.
!end if

if (testFlag) print *, '  aaa interpolatePsfCubeToCcd'

    call message(AppMsg, Verbose, "  Converting PSF cube to chip &
      &coordinates for this CCD:")

    if ( size(psfPixelCnrsPhi,1) .ne. size(psfSamplesCubeTelcoord,2)+1&
    .or. size(psfPixelCnrsPhi,2) .ne. size(psfSamplesCubeTelcoord,3)+1)&
call error('bug1',"Bug in module lccorr_corrector. Please contact the code developer.")

allocate(psfImgPixCrnrRaw(size(psfPixelCnrsPhi,1), size(psfPixelCnrsPhi,2)))
allocate(psfImgPixCrnrSng(size(psfPixelCnrsPhi,1), size(psfPixelCnrsPhi,2)))
allocate(isOffChip(size(psfPixelCnrsPhi,1), size(psfPixelCnrsPhi,2)))

    numEsamples = size(psfSamplesCubeTelcoord, 1)

    ! Find the positions of the PSF-array pixel corners in the chip frame:

    call message(AppMsg, Verbose, "    Converting PSF grid theta-phi &
      &values to RAWXY.")

if (testFlag) print *, '  bbb0'
!!!    call telToRaw(psfPixelCnrsThetaArcsec, psfPixelCnrsPhi&
!!!     , psfImgPixCrnrRaw%x, psfImgPixCrnrRaw%y, isOffChip)
call local_telToRawArray(psfPixelCnrsThetaArcsec, psfPixelCnrsPhi&
 , psfImgPixCrnrRaw%x, psfImgPixCrnrRaw%y, isOffChip)
if (testFlag) print *, '  bbb01'
if (testFlag) then
  print *, size(psfPixelCnrsThetaArcsec,1), size(psfPixelCnrsThetaArcsec,2)
  print *, size(psfPixelCnrsPhi,1), size(psfPixelCnrsPhi,2)
  print *, size(psfImgPixCrnrRaw%x,1), size(psfImgPixCrnrRaw%x,2)
  print *, size(psfImgPixCrnrRaw%y,1), size(psfImgPixCrnrRaw%y,2)
  print *, size(isOffChip,1), size(isOffChip,2)
end if
!!!call telToRawArray(psfPixelCnrsThetaArcsec, psfPixelCnrsPhi&
!!!     , psfImgPixCrnrRaw%x, psfImgPixCrnrRaw%y, isOffChip)
!allocate(dummyX(size(psfPixelCnrsPhi,1),size(psfPixelCnrsPhi,2)))
!allocate(dummyY(size(psfPixelCnrsPhi,1),size(psfPixelCnrsPhi,2)))
!dummyX = psfImgPixCrnrRaw%x
!dummyY = psfImgPixCrnrRaw%y
!call telToRawArray(psfPixelCnrsThetaArcsec, psfPixelCnrsPhi&
!     , dummyX, dummyY, isOffChip)
    ! in ssclib/coordinate.
if (testFlag) print *, '  bbb1'

    ! Now find out how big the chip-coord psf image array will have to be:

    minRawXi = nint(minval(psfImgPixCrnrRaw%x))
    maxRawXi = nint(maxval(psfImgPixCrnrRaw%x))
    minRawYi = nint(minval(psfImgPixCrnrRaw%y))
    maxRawYi = nint(maxval(psfImgPixCrnrRaw%y))

if (testFlag) print *, '  ccc'

    if (minRawXi > maxRawXi .or. minRawYi > maxRawYi) then
call error('bug2',"Bug in module lccorr_corrector. Please contact the code developer.")
!      allocate(psfSamplesCubeCcd(numEsamples, 0, 0))
!  return
    end if
if (testFlag) print *, '  ddd'

    allocate(psfSamplesCubeCcd(numEsamples&
      , minRawXi:maxRawXi, minRawYi:maxRawYi))
if (testFlag) print *, '  eee'

    if (present(doDummy)) then
if (testFlag) print *, '  fff'
      if (doDummy) then
if (testFlag) print *, '  ggg'
        call warning('dummyPsfCube', "This is a dummy routine! It doesn't &
          &return the real psfSamplesCubeCcd.")
if (testFlag) print *, '  hhh denom=', real((1+maxRawXi-minRawXi)*(1+maxRawYi-minRawYi))
        psfSamplesCubeCcd = 1.0&
          / real((1+maxRawXi-minRawXi)*(1+maxRawYi-minRawYi))
if (testFlag) print *, '  iii'
  return
      end if
    end if

    psfSamplesCubeCcd = 0.0

    if (   minRawXi > chipMaxRawX .or. maxRawXi < chipMinRawX&
      .or. minRawYi > chipMaxRawY .or. maxRawYi < chipMinRawY) then

      call message(AppMsg, Verbose, "    No significant overlap of PSF &
        &onto the chip. Calculate selection area only.")
  return
    end if

    ! Final step before the regridding is to normalize the corner coordinates
    ! to the pixel frame of psfSamplesCubeCcd:

!    psfImgPixCrnrRaw%x = psfImgPixCrnrRaw%x - real(minRawXi - 1)
!    psfImgPixCrnrRaw%y = psfImgPixCrnrRaw%y - real(minRawYi - 1)
psfImgPixCrnrSng%x = psfImgPixCrnrRaw%x - real(minRawXi - 1, single)
psfImgPixCrnrSng%y = psfImgPixCrnrRaw%y - real(minRawYi - 1, single)

    call message(AppMsg, Verbose, "    Rebinning PSF image pixel values to the &
      &chip grid.")

    ! Now let us regrid:
!    do eI = 1, numEsamples
!if (testFlag) then
!  print *, ''
!  print *, '    jjja eI=', eI
!end if

if (testFlag) print *, '    jjja'

call regridCartesian(psfImgPixCrnrSng, psfSamplesCubeTelcoord&
  , psfSamplesCubeCcd, foreignStatus, logical(testFlag, bool))
      ! in ssclib/regridders.
if (testFlag) print *, '    jjjb'

      if (foreignStatus < 0) then
print *, foreignStatus
call error('badRegridStatus',"The subroutine regrid returned with a bad &
  &status. You should contact the code developer.")
      end if
if (testFlag) print *, '    jjjc'

do eI = 1, numEsamples
      norm = sum(psfSamplesCubeCcd(eI,:,:))
if (testFlag) print *, '    jjjd'
if (norm <= 0) call error('bug3',"Bug in module lccorr_corrector. Please contact the code developer.")
if (testFlag) print *, '    jjje'
      psfSamplesCubeCcd(eI,:,:) = psfSamplesCubeCcd(eI,:,:) / norm
if (testFlag) then
  print *, '    jjjf'
  if (eI == 1) call dumpImageToFits(psfSamplesCubeCcd(eI,:,:), 'psf_chip.ds')
end if
    end do

if (testFlag) print *, '  zzz interpolatePsfCubeToCcd'

  end subroutine interpolatePsfCubeToCcd

  !---------------------------------------------------------------------
  subroutine local_telToRawVector(tanThetaArcsec, phi, xRaw, yRaw, isOffChip, ccdNum)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    real(double),  intent(in)            :: tanThetaArcsec(:),&
                                            phi(size(tanThetaArcsec))
    real(double),  intent(out)           :: xRaw(size(tanThetaArcsec)),&
                                            yRaw(size(tanThetaArcsec))
    logical(bool), intent(out)           :: isOffChip(size(tanThetaArcsec))
    integer(int8), intent(out), optional :: ccdNum(size(tanThetaArcsec))

    real(double), dimension(:), pointer :: cc2X, cc2Y, cc2Z, cc1X, cc1Y, cc1Z

    !-------------------------------------------------------------------
    ! Start

    call CAL_telCoordToCamCoord2(tanThetaArcsec, phi, cc2X, cc2Y, cc2Z)
    call CAL_camCoord2ToCamCoord1(cc2X, cc2Y, cc2Z, cc1X, cc1Y, cc1Z)
    call CAL_releaseMemory(cc2X, cc2Y, cc2Z)

    if (present(ccdNum)) then
      call local_cc1ToRawMasterCcd(cc1X, cc1Y, xRaw, yRaw, isOffChip, ccdNum)
    else
      call local_cc1ToRawMaster(cc1X, cc1Y, xRaw, yRaw, isOffChip)
    end if

    call CAL_releaseMemory(cc1X, cc1Y, cc1Z)

  end subroutine local_telToRawVector

  !---------------------------------------------------------------------
  subroutine local_telToRawArray(tanThetaArcsec, phi, xRaw, yRaw, isOffChip, ccdNum)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    real(double),  intent(in)            :: tanThetaArcsec(:,:),&
                                            phi(&
                                              size(tanThetaArcsec,1),&
                                              size(tanThetaArcsec,2))
    real(double),  intent(out)           :: xRaw(&
                                              size(tanThetaArcsec,1),&
                                              size(tanThetaArcsec,2)),&
                                            yRaw(&
                                              size(tanThetaArcsec,1),&
                                              size(tanThetaArcsec,2))
    logical(bool), intent(out)           :: isOffChip(&
                                              size(tanThetaArcsec,1),&
                                              size(tanThetaArcsec,2))
    integer(int8), intent(out), optional :: ccdNum(&
                                              size(tanThetaArcsec,1),&
                                              size(tanThetaArcsec,2))

    integer(int32) :: xSize, ySize, vSize
    real(double), allocatable ::&
      tanThetaArcsecVec(:), phiVec(:), xRawVec(:), yRawVec(:)
    logical(bool), allocatable ::&
      isOffChipVec(:)
    integer(int8), allocatable :: ccdNumVec(:)

    !-------------------------------------------------------------------
    ! Start

    xSize = size(tanThetaArcsec, 1)
    ySize = size(tanThetaArcsec, 2)
    vSize = xSize * ySize

    allocate(tanThetaArcsecVec(vSize), phiVec(vSize))
    allocate(xRawVec(vSize), yRawVec(vSize), isOffChipVec(vSize))
    tanThetaArcsecVec = reshape(tanThetaArcsec, (/vSize/))
    phiVec            = reshape(phi,            (/vSize/))

    if (present(ccdNum)) then
      allocate(ccdNumVec(vSize))
      call local_telToRawVector(tanThetaArcsecVec, phiVec, xRawVec, yRawVec&
        , isOffChipVec, ccdNumVec)

      ccdNum = reshape(ccdNumVec, (/xSize, ySize/))
    else
      call local_telToRawVector(tanThetaArcsecVec, phiVec, xRawVec, yRawVec&
        , isOffChipVec)
    end if

    xRaw      = reshape(xRawVec,      (/xSize, ySize/))
    yRaw      = reshape(yRawVec,      (/xSize, ySize/))
    isOffChip = reshape(isOffChipVec, (/xSize, ySize/))

  end subroutine local_telToRawArray

  !---------------------------------------------------------------------
  subroutine local_cc1ToRawMaster(xCc1, yCc1, xRaw, yRaw, isOffChip)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    real(double),  intent(in)  :: xCc1(:),&
                                  yCc1(size(xCc1))
    real(double),  intent(out) :: xRaw(size(xCc1)),&
                                  yRaw(size(xCc1))
    logical(bool), intent(out) :: isOffChip(size(xCc1))

    integer(int32) :: instrumentId
    integer(int16) :: chipXsize, chipYsize

    !-------------------------------------------------------------------
    ! Start

    !...................................................................
    ! Get max rawx and rawy and construct chip corners pseudo-vector:

    call CAL_getState(instrument=instrumentId)
    if (instrumentId == EPN) then
      chipXsize = EPN_MAX_X	! All these are defined in caloaldefs.
      chipYsize = EPN_MAX_Y

    else if (instrumentId == EMOS1 .or. instrumentId == EMOS2) then
      chipXsize = EMOS_MAX_X
      chipYsize = EMOS_MAX_Y

    else
call error('badInstrument','ssclib: Only EPIC instruments permitted.')
    end if

    call local_cc1ToRawSlave(chipXsize, chipYsize, xCc1, yCc1, xRaw, yRaw)

    ! Now check to see which values are off chip:
    isOffChip = .false. ! default
    where(nint(xRaw) < 1 .or. nint(xRaw) > chipXsize&
    .or.  nint(yRaw) < 1 .or. nint(yRaw) > chipYsize)&
      isOffChip = .true.

  end subroutine local_cc1ToRawMaster

  !---------------------------------------------------------------------
  subroutine local_cc1ToRawMasterCcd(xCc1, yCc1, xRaw, yRaw, isOffChip, ccdNums)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument.

    implicit none

    real(double),  intent(in)  :: xCc1(:),&
                                  yCc1(size(xCc1))
    real(double),  intent(out) :: xRaw(size(xCc1)),&
                                  yRaw(size(xCc1))
    logical(bool), intent(out) :: isOffChip(size(xCc1))
    integer(int8), intent(out) :: ccdNums(size(xCc1))

    real(double), allocatable ::&
      tempCc1X(:), tempCc1Y(:), tempCc1Z(:),&
!zCc1(:), 
      trialXRaw(:), trialYRaw(:), minXRaw(:), minYRaw(:)
    integer(int32) ::&
      instrumentId, numPoints, trialCcdNum, previousCcdNum, numCcds
    integer(int16) :: chipXsize, chipYsize
    integer(int16), pointer :: chipX(:), chipY(:)
    integer(int8), pointer :: onCcdNums(:)
    real(single), allocatable ::&
      minSeps(:), trialMinSeps(:)
    real(single) :: chipXCentre, chipYCentre
    integer(int8), allocatable :: minCcdNums(:)

    !-------------------------------------------------------------------
    ! Start

    numPoints = size(xCc1)

!    allocate(zCc1(numPoints))
!    zCc1 = 0.0d0
allocate(tempCc1X(numPoints), tempCc1Y(numPoints), tempCc1Z(numPoints))
tempCc1X = xCc1
tempCc1Y = yCc1
tempCc1Z = 0.0d0
!    call CAL_camCoord1ToChipCoord(xCc1, yCc1, zCc1, chipX, chipY, onCcdNums)
call CAL_camCoord1ToChipCoord(tempCc1X, tempCc1Y, tempCc1Z&
 , chipX, chipY, onCcdNums)

    call CAL_releaseMemory(chipX, chipY)

    ! Note that if position (xCc1, yCc1) does not lie on any CCD, the
    ! returned values of chipX, chipY and onCcdNums will all be zero.

    where(onCcdNums .ne. 0) isOffChip = .false.

    !...................................................................
    ! Get chip limits:

    call CAL_getState(instrument=instrumentId)
    if (instrumentId == EPN) then
      chipXsize = EPN_MAX_X	! All these are defined in caloaldefs.
      chipYsize = EPN_MAX_Y
      numCcds = 12

    else if (instrumentId == EMOS1 .or. instrumentId == EMOS2) then
      chipXsize = EMOS_MAX_X
      chipYsize = EMOS_MAX_Y
      numCcds = 7

    else
call error('badInstrument','ssclib: Only EPIC instruments permitted.')
    end if

    chipXCentre = (1 + chipXsize) / 2.0
    chipYCentre = (1 + chipYsize) / 2.0

    !...................................................................
    ! Now calculate RAWXY values and the CCDNR of the nearest chip:

    call CAL_getState(ccdChipId=previousCcdNum)

    trialCcdNum = 1
    call CAL_setState(ccdChipId=trialCcdNum)
!***** what about the node????


    allocate(minCcdNums(numPoints))
    allocate(trialXRaw(numPoints), trialYRaw(numPoints))
    allocate(minXRaw(numPoints), minYRaw(numPoints))
    allocate(minSeps(numPoints), trialMinSeps(numPoints))

    call local_cc1ToRawSlave(chipXsize, chipYsize, xCc1, yCc1, trialXRaw, trialYRaw)

    minSeps = max(abs(trialXRaw - chipXcentre)&
      , abs(trialYRaw - chipYcentre))

    where(.not. isOffChip .and. onCcdNums==1)
      xRaw = trialXRaw
      yRaw = trialYRaw
    elsewhere
      minXRaw = trialXRaw
      minYRaw = trialYRaw
      minCcdNums = 1
    end where

    do trialCcdNum = 2, numCcds
      call CAL_setState(ccdChipId=trialCcdNum)
      call local_cc1ToRawSlave(chipXsize, chipYsize, xCc1, yCc1, trialXRaw, trialYRaw)

      trialMinSeps = max(abs(trialXRaw - chipXcentre)&
        , abs(trialYRaw - chipYcentre))

      where(.not. isOffChip .and. onCcdNums==trialCcdNum)
        xRaw = trialXRaw
        yRaw = trialYRaw
      elsewhere(trialMinSeps < minSeps)
        minSeps = trialMinSeps
        minXRaw = trialXRaw
        minYRaw = trialYRaw
        minCcdNums = trialCcdNum
      end where
    end do

    where(isOffChip)
      xRaw = minXRaw
      yRaw = minYRaw
      ccdNums = minCcdNums
    elsewhere
      ccdNums = onCcdNums
    end where

    call CAL_setState(ccdChipId=previousCcdNum)

  end subroutine local_cc1ToRawMasterCcd

  !---------------------------------------------------------------------
  subroutine local_cc1ToRawSlave(chipXsize, chipYsize&
  , cc1X, cc1Y, trialXRaw, trialYRaw)
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument
  ! and ccdChipId and, if instrument is EMOS1 or EMOS2, for ccdNodeId too.

    implicit none

    integer(int16), intent(in) :: chipXsize, chipYsize
    real(double), intent(in) :: cc1X(:), cc1Y(size(cc1X))
!    type(Point2dDbleT), intent(out) :: trialRawReal(size(cc1X))
    real(double), intent(out) :: trialXRaw(size(cc1X)), trialYRaw(size(cc1X))

    real(double) :: mat(2,2), offsetCc1X, offsetCc1Y

    !-------------------------------------------------------------------
    ! Start

    !.................................................................
    ! Calculate the transform matrix and offset to enable the following
    ! transform:
    !
    !	( rawX )   ( mat(1,1)  mat(1,2) )   ( cc1X - offsetCc1X )
    !	(      ) = (                    ) * (                   )
    !	( rawY )   ( mat(2,1)  mat(2,2) )   ( cc1Y - offsetCc1Y )

    call local_calcTransformAndOffset(chipXsize, chipYsize&
      , mat, offsetCc1X, offsetCc1Y)

    !.................................................................
    ! Use this matrix to do all the conversions:

    call local_linearTransform(mat, cc1X, cc1Y, offsetCc1X, offsetCc1Y&
      , trialXRaw, trialYRaw)

  end subroutine local_cc1ToRawSlave


  !---------------------------------------------------------------------
  subroutine local_calcTransformAndOffset(chipXsize, chipYsize&
  , mat, offsetCc1X, offsetCc1Y)
  !---------------------------------------------------------------------
  ! The two big problems with the present call CAL_camCoord1ToChipCoord()
  ! are (i) the returned CHIPCOORD values are integer; (ii) off-chip
  ! positions are not defined. The next two subroutines are designed to
  ! work around this.
  !
  ! The present subroutine calculates the transform matrix and offset to
  ! enable the following transform:
  !
  !	( rawX )   ( mat(1,1)  mat(1,2) )   ( cc1X - offsetCc1X )
  !	(      ) = (                    ) * (                   )	(1)
  !	( rawY )   ( mat(2,1)  mat(2,2) )   ( cc1Y - offsetCc1Y )
  !
  ! This is done by calculating the elements of the inverse matrix tam()
  ! which satisfies the relation
  !
  !	( cc1X )   ( tam(1,1)  tam(1,2) )   ( rawX )   ( offsetCc1X )
  !	(      ) = (                    ) * (      ) + (            ),	(2)
  !	( cc1Y )   ( tam(2,1)  tam(2,2) )   ( rawY )   ( offsetCc1Y )
  !
  ! then by inverting to get mat() = tam()^-1.
  !
  ! The matrix tam() and the offset components are obtained as follows.
  ! Four pairs of rawX,Y values are taken from the corners of the CCD. The
  ! X and Y components are loaded into 4-element vectors for better ease
  ! of use of the CAL coordinate conversion routine(s). The order of the
  ! corners in these pseudo-vectors is as follows:
  !
  !	      -----------
  !	     |2         3|
  !	     |           |
  !	  ^  |           |
  !	  |  |           |
  !	R |  |           |
  !	A |  |1         4|
  !	W |   -----------
  !	Y |
  !	   -------->
  !	      RAWX
  !
  ! The CAL is used to convert these rawX,Y values to components in the
  ! CAMCOORD1 system. At the end of this step we have 4 sets of
  ! equation (2), one each for the four corners of the CCD.
  !
  ! Note now that if we subtract equation (2) for point 2 from that for
  ! point 3, then divide by (chipXsize - 1), the RHS is
  !
  !	( tam(1,1)  tam(1,2) )   ( 1 )   ( tam(1,1) )
  !	(                    ) * (   ) = (          ).	(3)
  !	( tam(2,1)  tam(2,2) )   ( 0 )   ( tam(2,1) )
  !
  ! Subtraction of equation (2) for point 1 from 4 gives the same result.
  ! Thus these two subtractions can be used to calculate tam(1,1) and
  ! tam(2,1) by averaging, viz:
  !
  !	(tam(1,1))   (cc1X(3)-cc1X(2)+cc1X(4)-cc1X(1))       0.5
  !	(        ) = (                               ) * -------------.	(4)
  !	(tam(2,1))   (cc1Y(3)-cc1Y(2)+cc1Y(4)-cc1Y(1))    chipXsize-1
  !
  ! (There is no real need to average, it just makes full (and
  ! symmetrical) use of the available data.) A similar equation yields
  ! the other two tam() components.
  !
  ! The offset vector can in principle be obtained from any of the four
  ! sets of corner values by doing
  !
  !	( offsetCc1X )   ( cc1X )   ( tam(1,1)  tam(1,2) )   ( rawX )
  !	(            ) = (      ) - (                    ) * (      ).	(5)
  !	( offsetCc1Y )   ( cc1Y )   ( tam(2,1)  tam(2,2) )   ( rawY )
  !
  ! However again it is good to average, hence in this routine the offset
  ! vector is computed for each of the four corners and the results averaged.
  !
  ! Note that the matrix indices are ordered in the (row, col) sense not the
  ! (x, y) sense.
  !
  !---------------------------------------------------------------------
  ! NOTE!! the cal must be set up by the calling routine for instrument,
  ! ccdChipId and (if instrument is EMOS1 or EMOS2) ccdNodeId.
  !---------------------------------------------------------------------

    implicit none

    integer(int16), intent(in) :: chipXsize, chipYsize
    real(double), intent(out) :: mat(2,2), offsetCc1X, offsetCc1Y

    integer(int16) ::&
      chipCornersRawX(4), chipCornersRawY(4), i
    real(double), pointer ::&
      chipCornersCc1X(:), chipCornersCc1Y(:), chipCornersCc1Z(:)
    real(double) :: tam(2,2), det
    logical :: oldRandomize

    !-------------------------------------------------------------------
    ! Start

    chipCornersRawX(1) = 1
    chipCornersRawX(2) = 1
    chipCornersRawX(3) = chipXsize
    chipCornersRawX(4) = chipXsize

    chipCornersRawY(1) = 1
    chipCornersRawY(2) = chipYsize
    chipCornersRawY(3) = chipYsize
    chipCornersRawY(4) = 1

    call CAL_getState(randomize=oldRandomize)
    call CAL_setState(randomize=.false.)
    call CAL_rawXY2mm(chipCornersRawX, chipCornersRawY&
      , chipCornersCc1X, chipCornersCc1Y, chipCornersCc1Z&
      , coordSysType=CAMCOORD1)

    call CAL_setState(randomize=oldRandomize)

    !...................................................................
    ! Set up CAMCOORD1 -> PIXCOORD linear transform matrix and offset vector:

    tam(1,1) = (chipCornersCc1X(3) - chipCornersCc1X(0)&
             +  chipCornersCc1X(2) - chipCornersCc1X(1))&
             * 0.5 / (real(chipXsize) - 1.0)

    tam(2,1) = (chipCornersCc1Y(3) - chipCornersCc1Y(0)&
             +  chipCornersCc1Y(2) - chipCornersCc1Y(1))&
             * 0.5 / (real(chipXsize) - 1.0)

    tam(1,2) = (chipCornersCc1X(1) - chipCornersCc1X(0)&
             +  chipCornersCc1X(2) - chipCornersCc1X(3))&
             * 0.5 / (real(chipYsize) - 1.0)

    tam(2,2) = (chipCornersCc1Y(1) - chipCornersCc1Y(0)&
             +  chipCornersCc1Y(2) - chipCornersCc1Y(3))&
             * 0.5 / (real(chipYsize) - 1.0)

    ! Invert:

    det = tam(1,1) * tam(2,2) - tam(1,2) * tam(2,1)
    mat(1,1) =  tam(2,2) / det
    mat(1,2) = -tam(1,2) / det
    mat(2,1) = -tam(2,1) / det
    mat(2,2) =  tam(1,1) / det

    !...................................................................
    ! Now calculate the offset vector:

    offsetCc1X = 0.0
    offsetCc1Y = 0.0
    do i = 1, 4
      offsetCc1X = offsetCc1X + chipCornersCc1X(i-1)&
        - (tam(1,1) * chipCornersRawX(i) + tam(1,2) * chipCornersRawY(i))
      offsetCc1Y = offsetCc1Y + chipCornersCc1Y(i-1)&
        - (tam(2,1) * chipCornersRawX(i) + tam(2,2) * chipCornersRawY(i))
    end do
    offsetCc1X = offsetCc1X / 4.0
    offsetCc1Y = offsetCc1Y / 4.0

    call CAL_releaseMemory(chipCornersCc1X, chipCornersCc1Y, chipCornersCc1Z)

  end subroutine local_calcTransformAndOffset

  !---------------------------------------------------------------------
  subroutine local_linearTransform(mat, cc1X, cc1Y, offsetCc1X, offsetCc1Y, xRaw, yRaw)
  !---------------------------------------------------------------------
  ! The two big problems with the present call CAL_camCoord1ToChipCoord()
  ! are (i) the returned CHIPCOORD values are integer; (ii) off-chip
  ! positions are not defined. The present and previous subroutines are
  ! designed to work around this.
  !
  ! The present subroutine calculates rawX,Y from the following relation:
  !
  !	( rawX )   ( mat(1,1)  mat(1,2) )   ( cc1X - offsetCc1X )
  !	(      ) = (                    ) * (                   )	(1)
  !	( rawY )   ( mat(2,1)  mat(2,2) )   ( cc1Y - offsetCc1Y )
  !

    implicit none

    real(double),       intent(in)  :: mat(2,2),&
                                       cc1X(:),&
                                       cc1Y(size(cc1X)),&
                                       offsetCc1X,&
                                       offsetCc1Y
    real(double), intent(out) :: xRaw(size(cc1X)), yRaw(size(cc1X))

    !-------------------------------------------------------------------
    ! Start

    xRaw = mat(1,1) * (cc1X - offsetCc1X) + mat(1,2) * (cc1Y - offsetCc1Y)
    yRaw = mat(2,1) * (cc1X - offsetCc1X) + mat(2,2) * (cc1Y - offsetCc1Y)

  end subroutine local_linearTransform




  !---------------------------------------------------------------------
  subroutine recordRateNullsAsGtis(outSetName, rateColName, outGtiTabName)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: outSetName, rateColName, outGtiTabName

    type(DataSetT) :: outSet
    type(TableT) :: rateTab, outGtiTab
    type(ColumnT) :: col
    integer :: numRows, row, numGtis, gtiNum, startRow, finisRow
    logical, allocatable :: rateIsNull(:)
    integer, allocatable :: startRows(:), finisRows(:)
    logical :: withinGti
    real(double), pointer :: binCentreTime(:), startVals(:), stopVals(:)
    real(single), pointer :: timeDel(:)
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    outSet = dataSet(outSetName, MODIFY)
    rateTab = table(outSet, 'RATE')
    numRows = numberOfRows(rateTab)

    binCentreTime => real64Data(column(rateTab, 'TIME'))

    allocate(timeDel(0:numRows-1))
    if (hasColumn(rateTab, 'TIMEDEL')) then
      call readColDataToFixed(rateTab, 'TIMEDEL', timeDel)
    else
      timeDel = real64Attribute(rateTab, 'TIMEDEL')
    end if

    col = column(rateTab, rateColName)
    allocate(rateIsNull(numRows))
    rateIsNull = .false. ! default
    do row = 0, numRows - 1
      if (isNull(col, row)) rateIsNull(row+1) = .true.
    end do

    ! Calculate the start and end rows of stretches of TRUE values:

    allocate(startRows(numRows), finisRows(numRows))
    startRows = 0
    finisRows = 0

if (testFlag) print *, ''
if (testFlag) print *, ''
    numGtis = 0
    withinGti = .false.
    do row = 1, numRows
      if (rateIsNull(row)) then
        if (withinGti) then
          finisRows(numGtis) = row - 1
          withinGti = .false.
if (testFlag) print *, '    rateIsNull,     withinGti. row=', row
        else
if (testFlag) print *, '    rateIsNull, NOT withinGti. row=', row
        end if
      else
        if (.not.withinGti) then
          numGtis = numGtis + 1
          startRows(numGtis) = row
          withinGti = .true.
if (testFlag) print *, 'NOT rateIsNull, NOT withinGti. row=', row
        else
if (testFlag) print *, 'NOT rateIsNull,     withinGti. row=', row
        end if
      end if
    end do
    if (withinGti) then
      finisRows(numGtis) = numRows
    end if
if (testFlag) print *, ''
if (testFlag) print *, ''

    outGtiTab = addTable(outSet, outGtiTabName, numGtis)
    startVals => real64Data(addColumn(outGtiTab, 'START', REAL64, 's'))
    stopVals  => real64Data(addColumn(outGtiTab, 'STOP',  REAL64, 's'))
    do gtiNum = 0, numGtis - 1
      startRow = startRows(gtiNum+1)-1
      finisRow = finisRows(gtiNum+1)-1
if (testFlag) print *, 'gtiNum, startRow, finisRow:', gtiNum+1, startRow+1, finisRow+1
      startVals(gtiNum) = binCentreTime(startRow) - timeDel(startRow) / 2.0d0
      stopVals(gtiNum)  = binCentreTime(finisRow) + timeDel(finisRow) / 2.0d0
    end do
if (testFlag) print *, ''
if (testFlag) print *, ''

    call setAttribute(block(outGtiTab), attribute(rateTab, 'MJDREF'))
    if (hasAttribute(rateTab, 'TIMEZERO')) then
      call setAttribute(block(outGtiTab), attribute(rateTab, 'TIMEZERO'))
    else
      call setAttribute(outGtiTab, 'TIMEZERO', 0.0d0, 's'&
        , "It's necessary to include this until SOC SPR-2565 is fixed.")
    end if

    call release(outSet)
    deallocate(timeDel)

  end subroutine recordRateNullsAsGtis

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module lccorr_corrector
