!
! Module: lccorr_pcms_mod
!
! Authors:  Ian Stewart - SSC
!          
! Description: This task calculates and optionally applies an exposure
! correction for timeseries.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
!
! ToDo:
!********** oote corrn not yet supported.
!	- nint() calls deprecated??
!	- look for submode keyword.
!	- doco: treatPatterns has no effect if treatQE=no.
!	- write also comments for *APP keywords.
!	- write bkg/src area ratio as a keyword?
!	- make the psf normalisation more elegant.
!	- if submode kwd exists, compare with mode worked out within lccorr.
!	- call CAL_releaseMemory when appropriate.
!	- Make test files on the fly! That will shrink the size of the upload by a large factor.
!	Change filter names and warn if filter='NONE' found - see UL emails.
!	If exposure keyword found in ts, overwrite with the time-average of fracexp. evselect writes a rough-and-ready exposure kwd but it deserves to be improved in lcurves that have been through lccorr.
!	Fix up column dimensions stuff - see runrgs on bryan.
!	Check more carefully to see that what happens when the bkg comes from BACKC etc columsn within the src TS is the same as when it comes from a separate file - eg, is it corrected for area? Subtracted if this is desired?
!	Check that, if bkg sub is desired but exp corr not, the bkg is divided by its own fracexp then multiplied by the source one. (Or maybe this is done when the areas are adjusted??)
!	Test for existence of DSS before attempting to open it. Not every TS gets region-filtered. Maybe the whole camera is desired!
!	Pixel areas need to be passed to sub lccorr by calcExposureCorrections. The purpose is not the same as the regionAreaMm2, which is needed in the calcSpatial routines.
!	However..... check this statement, 'the purpose is not the same' out. I have a feeling that it may be the same after all...
!	PN missing frames - see how this is handled.
!	If the mode (eg small window) does not allow a tally of all OOTEs, change the calc of dead time to take into account the region y-dimension for each column. This makes it more accurate.
!	Speed up function psf() by preloading coordinates referred to the psf map, instead of CC2 coords.
!	Change DIM of patternIsCanonical() and patternWasSelected() sometime from 1:32 to 0:31 (saves on lots of -1s in the CAL_getQM calls). NOTE! Tried this and ran into troubles.
!	Copy TIME column to output if it is present in both src and bkg TS. May need to check that all vals of the column are the same for src and bkg. (This may seem like an exacting criterion, but src and bkg should be made from the same evlist.)
!	Fix up messages to format numbers properly (without line breaks!)
!	Modify to cope with unequal-width time bins. Check out whether can enable ((TIME column) .and. (no TIMEDEL kwd)). Does the OGIP standard specify a TIMEDEL column in this case to take care of unequal-width bins?
!	Modify the time routines to use the TIME column where this is available, only calculating bin times from the bin number + srcTsTime1stBin where it is not available.
!	Carefully compare the GTI routine to the subsequent one - changes made to the latter which need to be aped by the former.
!	.Take out workarounds when JB fixes timeunits in evlists (+ exposu exts) from SECONDS to s.
!	Reinstate all the (i) eff area, (ii) psf, (iii) vignet stuff when the relevant CAL routines start working.
!	Clean up some of the -1s inserted when I found that the DAL and CAL return arrays starting from 0 not 1!! Bastards.
!	Empty ep frames - how to handle?
!	See DLumb's email of 17 Mar re integ time definition in the ccf.
!	Note in the doco that dss not copied over (because of confusion with bkg)
!	Use foreachsubtable when implemented.
!	Check that all opened objects are closed
!	Minimise memory consumption by not leaving objects open unless nec.
!	Think about timing modes, see if it makes sense to do exp corr for these, and if so which and how.
!	Loops that do more than 1 column at once: see if can break into >1 loop, each doing 1 column.
!	Mention in docs that burst, timing modes not yet supported.
!	Could be a problem if source GTI correction proceeds smoothly, but bkg doesn't. Better if the 2 TS are consistent in the corrections applied.
!	20040610: see if can speed up calcPnCcdExpMap by not opening the event set to get the badpix/offsets map for this ccd (which takes up most of the time) unless there are some selection-region pixels on the ccd.

module lccorr_pcms_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use cal		! for caltypes -> ModeParameterDataType
  use param  
  use errorHandling
  use dss
  use eventAttributes	! in saslib/sasdal. For checking which flags have
			! been applied.

  use lccorr_include
  use lccorr_corrector

!  use array_utils	! rebinVector()
  use cal_aux		! getCalInfo()
  use constants
  use coordinate	! for detUnitArcsec, detToRaDec, rawToDet, tanToRaDec
  use dal_aux		! readColDataToFixed()
  use dss_aux		! for calls in getCcdNumFromDssComp.
  use dss_bitmasks_aux
  use exposure_map_pcms	! for localTAWtype.
  use rebinners
  use ssc_misc		! fracDiff()
  use wcs_aux		! readWcs()

!use test_mod

  implicit none 

  private 

  public lccorr_pcms

  character(len=20) ::&
    timeUnit
  real(double) ::&
    mjdRef,&
    expStartTimeStamp
  type(SpacecraftAttitudeType) ::&
    scAttitude,&
    instrumAttitude
  type(WcsT) ::&
    wcs
  integer(int8) ::&
    numCcds = -1,&
    numNodes = -1,&
    ccd2ccdnr(12,2) = -1,&
    ccdnr2ccd(17) = -1,&
    ccdnr2node(17) = -1
  character(2) ::&
    ccdStr(12,2) = ''

contains 

  !---------------------------------------------------------------------
  subroutine lccorr_pcms
  !---------------------------------------------------------------------
  !
  ! Lccorr performs the following operations:
  !	(i) If the input source TS has integer COUNT values, these are
  ! converted to real-valued count rates by dividing by the bin duration
  ! and stored in a column RATE in the output TS. An ERROR column, containing
  ! uncertainties in the count rates, is also created appropriately.
  !	(ii) Exposure (+ oote & binning noise) corrections are calculated
  ! for the source TS. Most individual corrections are optional. The exposure
  ! factor per bin is written to a column FRACEXP in the output timeseries,
  ! whereas the oote and binning noise corrections are performed directly on
  ! the RATE values.
  !	(iii) Optionally, the exposure correction is performed by dividing
  ! the RATE column by the FRACEXP column.
  !	(iv) If a bkg TS is supplied, identical calculations are carried
  ! out on this TS. The options are shared with the source TS - eg it is not
  ! possible to correct the source TS for vignetting but not the background TS.
  !	(v) The bkg count rate values are normalised to the same collection
  ! area as the source. This is non-optional.
  !****** should be weighted by gtis, vignetting on each chip.
  !****** what about nonvignetted background!!??
  !	(vi) Optionally, the background is subtracted from the source.
  ! Before the above operations are performed, the relevant columns of the
  ! source and background TS are copied to the output TS and all the
  ! manipulations are performed on this output TS. When all operations are
  ! complete, the columns of the output TS are renamed and pruned as
  ! appropriate until it complies with the specification for the output of
  ! lccorr. At present this dictates that the output file has the same
  ! headers as the source TS, the same number and size of bins, and five
  ! real32-valued columns:
  !		RATE
  !		ERROR
  !		BACKC
  !		BACKE
  !		FRACEXP
  !......................................................................

    implicit none

    real(single) :: srcBkgWeight, bkgBkgWeight, ratioBkgOverSrcAreas
    integer(int32) :: requiredCorrections, filtId
    real(double) :: srcRaDeg, srcDecDeg
    type(DataSetT) :: outSet, set
    type(TableT) :: outTab, tab
    type(localTAWtype) :: timeAttWcs
    type(expInfoType), allocatable :: expInfo(:,:)
    real(single), allocatable :: ccdWeights(:,:)
    character(120) :: weightListStr
logical :: testFlag = .false.

    !--------------------------------------------------------------------
    ! Start:

if (testFlag) print *, 'aaa lccorr_pcms'
    call readParameters

if (testFlag) print *, 'bbb'
    requiredCorrections = 0 ! default
    if (treatFilterTrans)&
      requiredCorrections = ibset(requiredCorrections, TREAT_FILTER_BIT)
    if (treatVignet)&
      requiredCorrections = ibset(requiredCorrections, TREAT_VIGNET_BIT)
    if (treatQuantumEff)&
      requiredCorrections = ibset(requiredCorrections, TREAT_QE_BIT)
    if (treatGti)&
      requiredCorrections = ibset(requiredCorrections, TREAT_GTI_BIT)
    if (treatCosmicRays)&
      requiredCorrections = ibset(requiredCorrections, TREAT_COSMIC_BIT)
    if (treatDeadTime)&
      requiredCorrections = ibset(requiredCorrections, TREAT_DEAD_TIME_BIT)
    if (treatAlias)&
      requiredCorrections = ibset(requiredCorrections, TREAT_ALIAS_BIT)
!**** Neaten this up some time.
if (testFlag) print *, 'ccc'

    !....................................................................
    ! Open the source and bkg timeseries and perform some checks:

    call readTsKeywords(srcTsSetName, instrumentId, timeUnit, mjdRef)
    call checkTsHeader(srcTsSetName)

    if (withBkgTsSet) then
      call checkTsHeader(bkgTsSetName)
      call checkBkgAgainstSrc
    end if

    call constructOutSet

    call miscSetups ! sets the cal state for instrument in the call
    ! getCalInfo().
if (testFlag) print *, 'ddd'

!!!    call checkTsDss(srcTsSetName)
!!!    if (withBkgTsSet) call checkTsDss(bkgTsSetName)
if (testFlag) print *, 'eee'

    !....................................................................
    ! Open the event set and load some arrays from its header keywords and
    ! those of its EXPOSUnn extensions; load other miscellaneous parameters.

    allocate(expInfo(numCcds,numNodes), ccdWeights(numCcds,numNodes))
if (testFlag) print *, 'fff'
    call readCcdExposureInfo(eventSetName, ccdStr(:numCcds,:numNodes)&
      , instrumentId, expInfo, filtId) ! in exposure_map.
if (testFlag) print *, 'gg'

    set = dataSet(eventSetName, READ)
    tab = table(set, 'EVENTS')
    call readWcs(tab, 'X', 'Y', wcs) ! in ssclib/wcs_aux.
    call release(set)
if (testFlag) print *, 'hhh'

    call CAL_setState(filterId=filtId)
    timeAttWcs%timeStamp = expStartTimeStamp
    timeAttWcs%attitude  = scAttitude
    timeAttWcs%wcs       = wcs
if (testFlag) print *, 'iii'

    !....................................................................
    ! Perform the required exposure corrections on the source dataset:

    call message(AppMsg, Verbose, "########################################")
    call message(AppMsg, verbose, 'Calculating exposure corrections &
      &for the source region.')
if (testFlag) print *, 'jjj'

    ccdWeights = 0.0 ! default
    select case(srcWeightStyle)
    case('none')
      ccdWeights(:,1) = 1.0
      srcBkgWeight = 1.0
    case('user')
      if (size(userSrcWeightListNode0) .ne. numCcds)&
call error('wrongNumberUserSrcWeights',"The number of elements in the &
&list parameter --srcweightsnode0 must match the number of CCDs.")

      ccdWeights(:,1) = userSrcWeightListNode0
      if (withSrcNode1) then
        if (numNodes <= 1) then
          call warning('noRedundantNodeThisInstrument',"")
        else
          ccdWeights(:,2) = userSrcWeightListNode1
        end if
      end if

      srcBkgWeight = userSrcRegionArea

if (all(ccdWeights <= 0.0)) call error('allWeightsZero',"At least 1 of &
&the elements supplied to --srcweightsnode0 or --srcweightsnode1 must be &
&greater than zero.")

      ccdWeights = ccdWeights / sum(ccdWeights)

    case('deduce')
      call getSrcPositions(wcs, srcRaDeg, srcDecDeg)
      ! this needs the evlist wcs.

!*** write src pos to output (and also whether this was obtained from the dss or from the user??)

      select case(srcWeightDeductionStyle)
      case('psf')
        call calcCcdWeightsFromTheory(srcTsSetName, timeAttWcs, expInfo&
          , requiredCorrections, ccdWeights, srcBkgWeight, srcRaDeg, srcDecDeg)

        ! This if-block is duplicated in the following case('events')
        if (srcBkgWeight <= 0.0) then
           call warning('noSrcArea',"The task found that no &
                &events from the source data set passed the selection &
                &expression.  Exit with no product created.")
           return		! File tempOutSetName (='tempset') left
        end if

if (all(ccdWeights <= 0.0)) call error('allSpatialMapsEmpty',"None of the &
  &CCD maps of source spatial selections contain any non-zero pixels.")

        ccdWeights = ccdWeights / sum(ccdWeights)

      case('image')
call error('notYetSupported',"")

      case('events')
        call calcCcdWeightsFromEvlist(srcTsSetName, timeAttWcs, expInfo&
          , ccdWeights, srcBkgWeight, srcRaDeg, srcDecDeg)

        ! This if-block is duplicated in the previous case('psf')
        if (srcBkgWeight <= 0.0) then
           call warning('noSrcArea',"The task found that no &
                &events from the source data set passed the selection &
                &expression.  Exit with no product created.")
           return		! File tempOutSetName (='tempset') left
        end if

if (all(ccdWeights <= 0.0)) call error('filteredEventListEmpty',"No events &
  &could be found after filtering the event list on the source extraction region.")

      ccdWeights = ccdWeights / sum(ccdWeights)

      case default
call error('badSrcWeightDeductionStyle',"")
      end select

    case default
call error('badSrcWeightStyle',"")
    end select
if (testFlag) print *, 'kkk'

    write(errStr, '(a, es10.4)') "Source-region pixel area is ", srcBkgWeight
    call message(AppMsg, VERBOSE)

    ! Note: ccdWeights are (in general) unnormalized at this point.
    ! The normalization is performed inside calcExposureCorrections.

    call calcWeightListStr(ccdWeights(:,1), weightListStr)
    if (numNodes > 1) then
      call message(AppMsg, VERBOSE, "Source weight list (primary node):   '"&
        //trim(weightListStr)//"'")
      call calcWeightListStr(ccdWeights(:,2), weightListStr)
      call message(AppMsg, VERBOSE, "Source weight list (redundant node): '"&
        //trim(weightListStr)//"'")
    else
      call message(AppMsg, VERBOSE, "Source weight list: '"&
        //trim(weightListStr)//"'")
    end if

    call calcExposureCorrections(&
      srcTsSetName,&
      tempOutSetName,&
      eventSetName,&
      expInfo,&
      requiredCorrections,&
      .true.,&
      applyCorrections,&
      ccdWeights&
    )

    call recordRateNullsAsGtis(tempOutSetName, 'RATE', 'SRC_GTIS')

!!    outSet = dataSet(outSetName, MODIFY)
!outSet = dataSet(tempOutSetName, MODIFY)
!    outTab = table(outSet, 'RATE')
!    call setAttribute(outTab, 'EFFAREA', srcWeightedEffAreaCm2&
!      , comment='Mirror effective area in cm^2, weighted by spectrum.')
!    call setAttribute(outTab, 'SKYAREA', srcRegionAreaSr&
!      , comment='Area in steradians of collection region')

if (testFlag) print *, 'lll'
    !....................................................................
    !....................................................................

    if (withBkgTsSet) then
      ! Perform the required corrections on the background timeseries:

      call message(AppMsg, Verbose, "########################################")
      call message(AppMsg, verbose, 'Calculating exposure corrections &
        &for the background region.')

      ccdWeights = 0.0 ! default
      select case(bkgWeightStyle)
      case('none')
        ccdWeights(:,1) = 1.0
        bkgBkgWeight = 1.0
      case('user')
        if (size(userBkgWeightListNode0) .ne. numCcds)&
call error('wrongNumberUserBkgWeights',"The number of elements in the &
&list parameter --bkgweightsnode0 must match the number of CCDs.")

        ccdWeights(:,1) = userBkgWeightListNode0
        if (withBkgNode1) then
          if (numNodes <= 1) then
            call warning('noRedundantNodeThisInstrument',"")
          else
            ccdWeights(:,2) = userBkgWeightListNode1
          end if
        end if

        bkgBkgWeight = userBkgRegionArea

        if (all(ccdWeights <= 0.0)) then
          call warning('allWeightsZero',"At least 1 of &
            &the elements supplied to --bkgweightsnode0 or &
            &--bkgweightsnode1 should be greater than zero.")
        else
          ccdWeights = ccdWeights / sum(ccdWeights)
        end if

      case('flat')
        call calcCcdWeightsFromTheory(bkgTsSetName, timeAttWcs, expInfo&
          , requiredCorrections, ccdWeights, bkgBkgWeight)

        if (bkgBkgWeight <= 0.0) call warning('noBkgArea',"The task found &
          &that no events from the background data set passed the selection &
          &expression.")

        if (all(ccdWeights <= 0.0)) then
          call warning('allSpatialMapsEmpty',"None of the maps of &
            &background spatial selections contained non-zero pixels &
            &for any CCD.")
        else
          ccdWeights = ccdWeights / sum(ccdWeights)
        end if

      case('image')
call error('notYetSupported',"")
      case('events')
call error('notYetSupported',"")
      case default
call error('badBkgWeightStyle',"")
      end select

      write(errStr, '(a, es10.4)') "Background-region pixel area is ", bkgBkgWeight
      call message(AppMsg, VERBOSE)

      ! Note: ccdWeights are (in general) unnormalized at this point.
      ! The normalization is performed inside calcExposureCorrections.

      call calcWeightListStr(ccdWeights(:,1), weightListStr)
      if (numNodes > 1) then
        call message(AppMsg, VERBOSE, "Background weight list (primary &
          &node):   '"&
          //trim(weightListStr)//"'")
        call calcWeightListStr(ccdWeights(:,2), weightListStr)
        call message(AppMsg, VERBOSE, "Background weight list (redundant &
          &node): '"//trim(weightListStr)//"'")
      else
        call message(AppMsg, VERBOSE, "Background weight list: '"&
          //trim(weightListStr)//"'")
      end if

      call calcExposureCorrections(&
        bkgTsSetName,&
        tempOutSetName,&
        eventSetName,&
        expInfo,&
        requiredCorrections,&
        .false.,&
        applyCorrections,&
        ccdWeights&
      )

      call recordRateNullsAsGtis(tempOutSetName, 'BRATE', 'BKG_GTIS')

      ratioBkgOverSrcAreas = bkgBkgWeight/srcBkgWeight

      write(errStr, '(a, es10.4)') "Ratio of background to source &
        &collection areas is ", ratioBkgOverSrcAreas
      call message(AppMsg, VERBOSE)

      outSet = dataSet(tempOutSetName, MODIFY)
      outTab = table(outSet, 'RATE')
      call setAttribute(outTab, 'BKGRATIO', ratioBkgOverSrcAreas, '')
      call rescaleBkg(outTab, ratioBkgOverSrcAreas)

      if (subtractBkg) then
        call subtractCompensatedBkg(outTab)
        call setAttribute(outTab, 'BACKAPP', .true., '')
        call setAttribute(outTab, 'HDUCLAS2', 'NET'&
          , 'Background has been subtracted from source count rates.')

      else
        call setAttribute(outTab, 'BACKAPP', .false., '')
        call setAttribute(outTab, 'HDUCLAS2', 'TOTAL'&
          , 'Background has not been subtracted from source count rates.')
      end if
      call release(outSet)
    end if ! withBkgTsSet.
if (testFlag) print *, 'mmm'

    ! Up until now, all possible information has been preserved - counts,
    ! rate, error in counts and rate, and fracexp, for both of source and
    ! background if applicable (although the correction for collection area
    ! and the background subtraction is at present only applied to RATE,
    ! RERROR (-> ERROR in the output file), BRATE (-> BACKV) and BRERROR
    ! (-> BACKE). Now these columns are pruned to just that set desired in
    ! the output. If the specification for the product changes at a later
    ! date, such a change can be implemented reasonably simply by changing
    ! the routine below.

    call finaliseOutSet

if (testFlag) print *, 'zzz lccorr_pcms'
  end subroutine lccorr_pcms

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: numNode0Weights, weightNum, i, gtiNum

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Getting command-line parameters.')

    srcTsSetName = stringParameter('srctsset')
    eventSetName = stringParameter('eventset')

    withBkgTsSet = booleanParameter('withbkgtsset')
    if (withBkgTsSet) then
      bkgTsSetName = stringParameter('bkgtsset')
      bkgWeightStyle = stringParameter('bkgweightstyle')
      select case(bkgWeightStyle)
      case('none')
      case('user')
        numNode0Weights = parameterCount('bkgweightsnode0')
        allocate(userBkgWeightListNode0(numNode0Weights))
        do weightNum = 1, numNode0Weights
          userBkgWeightListNode0(weightNum) = realParameter('bkgweightsnode0'&
            , weightNum-1)
        end do

        withBkgNode1 = booleanParameter('withbkgnode1')
        if (withBkgNode1) then
          if (parameterCount('bkgweightsnode1') .ne. numNode0Weights)&
call error('wrongNumBkgNode1Weights',"The number of elements in list &
&parameter --bkgweightsnode1 must equal the number in --bkgweightsnode0.")
          allocate(userBkgWeightListNode1(numNode0Weights))
          do weightNum = 1, numNode0Weights
            userBkgWeightListNode1(weightNum) = realParameter('bkgweightsnode1'&
              , weightNum-1)
          end do
        end if

        userBkgRegionArea = realParameter('bkgarea')

      case('flat')
      case('image')
call error('notYetSupported',"")
!        bkgImgSetName = stringParameter('bkgimgset')
      case('events')
call error('notYetSupported',"")
      case default
call error('badBkgWeightStyle',"The value of --bkgweightstyle was not recognized")
      end select

!      withBkgSpecSet = booleanParameter('withbkgspecset')
!      bkgSpecSetName = stringParameter('bkgspecset')

      subtractBkg = booleanParameter('subtractbkg')
      bkgRegionTabName = stringParameter('bkgregiontab')
    end if

!    withSrcSpecSet = booleanParameter('withsrcspecset')
!    srcSpecSetName = stringParameter('srcspecset')

!    rebinSpec = booleanParameter('rebinspec')
!    rebinSpecMode = stringParameter('rebinspecmode')
!    numRBspecRows = intParameter('nspecbins')
!    specBinDiv = realParameter('specbindiv')

    srcWeightStyle = stringParameter('srcweightstyle')
    select case(srcWeightStyle)
    case('none')
    case('user')
      numNode0Weights = parameterCount('srcweightsnode0')
      allocate(userSrcWeightListNode0(numNode0Weights))
      do weightNum = 1, numNode0Weights
        userSrcWeightListNode0(weightNum) = realParameter('srcweightsnode0'&
          , weightNum-1)
      end do

      withSrcNode1 = booleanParameter('withsrcnode1')
      if (withSrcNode1) then
        if (parameterCount('srcweightsnode1') .ne. numNode0Weights)&
call error('wrongNumSrcNode1Weights',"The number of elements in list &
&parameter --srcweightsnode1 must equal the number in --srcweightsnode0.")
        allocate(userSrcWeightListNode1(numNode0Weights))
        do weightNum = 1, numNode0Weights
          userSrcWeightListNode1(weightNum) = realParameter('srcweightsnode1'&
            , weightNum-1)
        end do
      end if

      userSrcRegionArea = realParameter('srcarea')

    case('deduce')
      srcPositionStyle = stringParameter('srcposstyle')
      select case(srcPositionStyle)
      case('user')
        userSrcRaDeg  = realParameter('srcra')
        userSrcDecDeg = realParameter('srcdec')
      case('keyword')
        srcPositionSetName = stringParameter('srcposset')
      case('dss')
      case default
call error('badSrcPosStyle',"The value of --srcposstyle was not recognized.")
      end select

      srcWeightDeductionStyle = stringParameter('srcdeducestyle')
      select case(srcWeightDeductionStyle)
      case('psf')
        lowEnergyPsfEnergyFrac = realParameter('efrac')

      case('image')
call error('notYetSupported',"")
!        srcImgSetName = stringParameter('srcimgset')
      case('events')
      case default
call error('badSrcWeightDeductionStyle',"The value of --srcdeducestyle was not recognized.")
      end select

    case default
call error('badSrcWeightStyle',"The value of --srcweightstyle was not recognized.")
    end select

    outSetName = stringParameter('outset')
    tempOutSetName = stringParameter('tempset')	! left in an problematic exit
    tempEventListSetName = stringParameter('tempeventset')

    treatVignet = booleanParameter('treatvignet')
    treatFilterTrans = booleanParameter('treatfiltertrans')
    treatQuantumEff = booleanParameter('treatquantumeff')
    treatCosmicRays = booleanParameter('treatcosmicrays')
    treatDeadTime = booleanParameter('treatdeadtime')
    treatGti = booleanParameter('treatgti')
    treatAlias = booleanParameter('treatalias')
!    treatOotes = booleanParameter('treatootes')  

    applyCorrections = booleanParameter('applycorrections')

! NOTE!!! Rebinning of time-dependent factors is currently not done. With
! the present algorithms for calculating fracexp, little or no time would
! be saved by rebinning.
!    nTimeBins = 0
!    timeBinDiv = 1.0
!    rebinTime = booleanParameter('rebintime')
!    if (rebinTime) then
!      rebinTimeMode = stringParameter('rebintimemode')
!      if (trim(rebinTimeMode) .eq. 'tonumber') then
!        nTimeBins = intParameter('ntimebins')
!      else
!        timeBinDiv = realParameter('timebindiv')
!      end if
!    end if



patternLo = intParameter('patternlo')
patternHi = intParameter('patternhi')
flagMaskHexStr = stringParameter('flagmask')
numUserEeVRanges = parameterCount('pilos')
if (numUserEeVRanges <= 0) call error('',"")
if (numUserEeVRanges .ne. parameterCount('pihis')) call error('',"")
allocate(userEeVRanges(numUserEeVRanges))
userEeVRanges%lower%type = INCLUSIVE
userEeVRanges%upper%type = INCLUSIVE
do i = 1, numUserEeVRanges
  userEeVRanges(i)%lower%value = intParameter('pilos', i-1)
  userEeVRanges(i)%upper%value = intParameter('pihis', i-1)
end do

withGtis = booleanParameter('withgtis')
if (withGtis) then
        numGtiTabs = parameterCount('gtitabsnode0')
        withNode1 = booleanParameter('withgtitabsnode1')
        if (withNode1) then
          if (parameterCount('gtitabsnode1') .ne. numGtiTabs)&
call error('wrongNumGtiTabsNode1',"The number of elements in list &
&parameter --gtitabsnode1 must equal the number in --gtitabsnode0.")
          allocate(userGtiSetTabNames(numGtiTabs, 2))
        else
          allocate(userGtiSetTabNames(numGtiTabs, 1))
        end if

        do gtiNum = 1, numGtiTabs
          userGtiSetTabNames(gtiNum, 1) = stringParameter('gtitabsnode0'&
            , gtiNum-1)
        end do

        if (withNode1) then
          do gtiNum = 1, numGtiTabs
            userGtiSetTabNames(gtiNum, 2) = stringParameter('gtitabsnode1'&
              , gtiNum-1)
          end do
        end if
end if
rawYLo = intParameter('rawylo')
srcRegionTabName = stringParameter('srcregiontab')


  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine readTsKeywords(tsSetName, tsInstrumentId, tsTimeUnit, tsMjdRef)
  !---------------------------------------------------------------------

    implicit none 

    character(*),   intent(in)  :: tsSetName
    integer(int32), intent(out) :: tsInstrumentId
    character(*),   intent(out) :: tsTimeUnit
    real(double),   intent(out) :: tsMjdRef

    type(DataSetT) ::&
      set
    type(TableT) ::&
      tab
    character(20) :: instrumentStr

    !-------------------------------------------------------------------
    ! Start:

!    set = dataSet(srcTsSetName, READ)
set = dataSet(tsSetName, READ)
    instrumentStr = stringAttribute(set, 'INSTRUME')
    tsInstrumentId = instId(trim(instrumentStr))

    if (tsInstrumentId .ne. EMOS1&
    .and. tsInstrumentId .ne. EMOS2&
    .and. tsInstrumentId .ne. EPN) then
      write(errstr, *) "Instrument "//trim(instrumentStr)//&
      " not supported."
call error('badInstrument')
    end if

    tab = table(set, 'RATE')
    tsTimeUnit = stringAttribute(tab, 'TIMEUNIT')
    tsMjdRef   = real64Attribute(tab, 'MJDREF')

call release(set)

  end subroutine readTsKeywords

  !---------------------------------------------------------------------
  subroutine checkTsHeader(tsSetName)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: tsSetName

    type(DataSetT) ::&
      tsSet
    type(TableT) ::&
      tsTab
    type(ColumnT) ::&
      timeCol
    character(len=20) ::&
      timeColUnits, tsTimeUnit

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, "Checking headers of "//trim(tsSetName))

    tsSet = dataSet(tsSetName, READ)
    tsTab = table(tsSet, 'RATE')

    if (numberOfRows(tsTab) .le. 0) then 
      write(errstr, *) "The timeseries "//trim(tsSetName)//" contains no data."
call error('tsNoData')
    end if

    if (hasAttribute(tsTab, 'NUMBAND')) then
      if (int16Attribute(tsTab, 'NUMBAND') .ne. 1) then
        write(errstr, *) "The timeseries "//trim(tsSetName)//" contains &
        &data in more than 1 energy band."
call error('tsTooManyEbands')
      end if
    end if

    tsTimeUnit = stringAttribute(tsTab, 'TIMEUNIT')

    if (hasColumn(tsTab, 'TIME')) then
      timeCol = column(tsTab, 'TIME')
      timeColUnits = units(timeCol)

! The following is a workaround until JB fixes emevents.
if (trim(timeColUnits) .eq. 'SECONDS') then
  call warning('deprecatedUnit', "OGIP standards require that the TIME column have units 's' not 'SECONDS'.")
  timeColUnits = 's'
end if
      if (trim(timeColUnits) .ne. ''&
     .and. trim(timeColUnits) .ne. trim(tsTimeUnit)) then
        write(errstr, *) "The timeseries has inconsistent time units."
call error('tsBadTimeUnits')
      end if

    else if (.not. hasAttribute(tsTab, 'TIMEDEL')) then
      write(errstr, *) "The timeseries has neither a TIME &
      &column nor a TIMEDEL keyword."
call error('tsNoTimes')
    end if

    if (withBkgTsSet .and. hasColumn(tsTab, 'BACKV')) then
      write(errstr, *) "The timeseries already has a BACKV &
      &column. This data will be discarded."

      call warning('tsHasBackv')
    end if

    call release(tsSet)

  end subroutine checkTsHeader

  !---------------------------------------------------------------------
  subroutine checkBkgAgainstSrc
  !---------------------------------------------------------------------

    implicit none 
        
    integer(int32) ::&
      bkgInstrumentId
    character(20) ::&
      bkgTimeUnit
    real(double) ::&
      bkgMJDref

    !-------------------------------------------------------------------
    ! Start:

    call readTsKeywords(bkgTsSetName, bkgInstrumentId, bkgTimeUnit, bkgMjdRef)

    if (bkgInstrumentId .ne. instrumentId) then
      write(errstr, *) "Source and background timeseries &
      &INSTRUME keywords don't match."
call error('tsDifferentInstrumeKwds')
    end if

    if (trim(bkgTimeUnit) .ne. trim(timeUnit)) then
      write(errstr, *) "Source and background timeseries &
      &TIMEUNIT keywords don't match."
call error('tsDifferentTimeunitKwds')
    end if

    if (bkgMjdRef .ne. mjdRef) then
      write(errstr, *) "Source and background timeseries &
      &MJDREF keywords don't match."
call error('tsDifferentMJDrefKwds')
    end if

  end subroutine checkBkgAgainstSrc

  !---------------------------------------------------------------------
  subroutine constructOutSet
  !---------------------------------------------------------------------
  ! Constructs the output dataset and fills its columns. The source and
  ! background time series won't be needed after this step except for
  ! dss access.

    implicit none

    type(DataSetT) :: srcTsSet, outSet, bkgTsSet
    type(TableT)   :: srcTsTab, outTab, bkgTsTab
    type(ColumnT) ::&
      outTimeCol, outTimeDelCol, col, outRateCol, outErrorCol,&
      outBkgRateCol, outBkgErrorCol
    type(DssT) ::&
      dssPointer
    integer(int32) :: numSrcTsBins, row
    real(double), pointer ::&
      srcTsTime(:),&
      outTime(:),&
      bkgTsTime(:),&
      srcTsBinEdges(:),&
      bkgTsBinEdges(:)
    real(double), allocatable ::&
      tempOutBkgRate(:),&
      tempOutBkgError(:)
    real(single), pointer ::&
      srcTsTimeDel(:),&
      outTimeDel(:),&
      outRate(:),&
      outError(:),&
      srcTsRate(:),&
      srcTsError(:),&
      outBkgRate(:),&
      outBkgError(:),&
      bkgTsTimeDel(:),&
      bkgTsRate(:),&
      bkgTsError(:)
    logical :: rebinningIsNecessary
    logical(bool), allocatable :: nulls(:)
    integer :: status
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    !....................................................................
    ! Create the output data set and copy over header keywords from the
    ! source dataset:

    call message(AppMsg, Verbose, 'Copying keywords and columns to output TS.')

    srcTsSet = dataSet(srcTsSetName, READ)
    srcTsTab = table(srcTsSet, 'RATE')
    numSrcTsBins = numberOfRows(srcTsTab)

!    outSet = dataSet(outSetName, CREATE)
outSet = dataSet(tempOutSetName, CREATE)	! left in an problematic exit
    call setAttributes(attributable(outSet), attributable(srcTsSet))
    outTab = addTable(outSet, 'RATE', numSrcTsBins)
    call setAttributes(attributable(outTab), attributable(srcTsTab))
    if (subtractBkg) call deleteAttribute(outTab, 'HDUCLAS2')
    call deleteAttribute(outTab, 'HDUCLAS3')
    call setAttribute(outTab, 'HDUCLAS3', 'RATE'&
    , 'Data is in the form of counts per sec.')

    if (applyCorrections) then
      call setAttribute(outTab, 'DEADAPP', .true., '')
      call setAttribute(outTab, 'VIGNAPP', .true., '')
    else
      call setAttribute(outTab, 'DEADAPP', .false., '')
      call setAttribute(outTab, 'VIGNAPP', .false., '')
    end if

if (testFlag) then
  if (hasAttribute(outTab, 'EXPOSURE')) then
    print *, 'has EXPOSURE before DSS delete'
  else
    print *, 'no EXPOSURE kwd'
  end if
end if

    call dssDelete(block(outSet, 'RATE')) ! Gets rid of DSS attributes
    ! carried over from the source RATE header.

if (testFlag) then
  if (hasAttribute(outTab, 'EXPOSURE')) then
    print *, 'has EXPOSURE after DSS delete'
  else
    print *, 'no EXPOSURE kwd'
  end if
end if

    if (.not. withBkgTsSet) then ! we can copy the src ts DSS
    ! to the output file.
      dssPointer = dataSubspace(block(srcTsSet, 'RATE'))
      call dssWrite(dssPointer, block(outSet, 'RATE'))
      call dssRelease(dssPointer)
    end if

    !....................................................................
    ! Fill time and fracexp columns:

    call message(AppMsg, Noisy, 'Fetching timing info from source TS table.')

    call fetchTimeAndDel(srcTsTab, srcTsTime, srcTsTimeDel)

    outTimeCol    = addColumn(outTab, 'TIME',    REAL64, 's')
    outTime    => real64Data(outTimeCol)
    outTime    = srcTsTime
    call release(outTimeCol)

    if (size(srcTsTimeDel) < 1)&
call error('noTimeDel',"No TIMEDEL information was found in the file.")

    if (size(srcTsTimeDel) == 1) then
      call setAttribute(outTab, 'TIMEDEL', srcTsTimeDel(1), 's')
    else
      if (all(srcTsTimeDel(2:) == srcTsTimeDel(:size(srcTsTimeDel)-1))) then
        call setAttribute(outTab, 'TIMEDEL', srcTsTimeDel(1), 's')
      else
        outTimeDelCol = addColumn(outTab, 'TIMEDEL', REAL32, 's')
        outTimeDel => real32Data(outTimeDelCol)
        outTimeDel = srcTsTimeDel
        call release(outTimeDelCol)
      end if
    end if

    col = addColumn(outTab, 'FRACEXP',  REAL32)
    call release(col)
    col = addColumn(outTab, 'BFRACEXP', REAL32)
    call release(col)

    !....................................................................
    ! Fill source rate + error columns:

    outRateCol  = addColumn(outTab, 'RATE',  Real32, 'counts/s')
    outErrorCol = addColumn(outTab, 'ERROR', Real32, 'counts/s')
    outRate  => real32Data(outRateCol)
    outError => real32Data(outErrorCol)

    call message(AppMsg, Noisy, 'Fetching rate and error from source TS table.')

    call fetchRateAndError(srcTsTab, srcTsTimeDel, srcTsSetName, srcTsRate&
      , srcTsError)
    outRate  = srcTsRate
    outError = srcTsError

    call release(outRateCol)
    call release(outErrorCol)
    call release(srcTsSet)
    deallocate(srcTsRate, srcTsError)

    !....................................................................
    ! Fill background rate + error columns:

!    outBkgRateCol  = addColumn(outTab, 'BRATE',  Real32, 'counts/s')
!    outBkgErrorCol = addColumn(outTab, 'BERROR', Real32, 'counts/s')
!    outBkgRate  => real32Data(outBkgRateCol)
!    outBkgError => real32Data(outBkgErrorCol)

    if (.not. withBkgTsSet) then
!      outBkgRate  = 0.0
!      outBkgError = 0.0
!****** this zeroing is done because, if no bkg ts is supplied, at present these columns remain in the output and are not filled. They probably should in fact be deleted from the output.

!*** In fact I have changed the code so that these columns are not added unless withBkgTsSet is true. After the doco is changed to reflect the now optional nature of these columns in the output, you can remove all the commented lines. IMS 20040213.

    else
outBkgRateCol  = addColumn(outTab, 'BRATE',  Real32, 'counts/s')
outBkgErrorCol = addColumn(outTab, 'BERROR', Real32, 'counts/s')
outBkgRate  => real32Data(outBkgRateCol)
outBkgError => real32Data(outBkgErrorCol)

      bkgTsSet = dataSet(bkgTsSetName, READ)
      bkgTsTab = table(bkgTsSet, 'RATE')

      call message(AppMsg, Noisy, 'Fetching timing info from background &
        &TS table.')

      call fetchTimeAndDel(bkgTsTab, bkgTsTime, bkgTsTimeDel)

      call message(AppMsg, Noisy, 'Fetching rate and error from background &
        &TS table.')

      call fetchRateAndError(bkgTsTab, bkgTsTimeDel&
      , bkgTsSetName, bkgTsRate, bkgTsError)

      rebinningIsNecessary = .false.
      if (size(srcTsTime) .ne. size(bkgTsTime)) then
        rebinningIsNecessary = .true.
      else if (any(srcTsTime .ne. bkgTsTime)) then
        rebinningIsNecessary = .true.
      end if

      if (rebinningIsNecessary) then
        call calcTimeBinEdges(srcTsTime, srcTsTimeDel, srcTsBinEdges)
        call calcTimeBinEdges(bkgTsTime, bkgTsTimeDel, bkgTsBinEdges)

        deallocate(bkgTsTime, bkgTsTimeDel)
        allocate(nulls(numSrcTsBins))
        allocate(tempOutBkgRate(numSrcTsBins))
        allocate(tempOutBkgError(numSrcTsBins))
if (testFlag) print *, '**** just before rebinVector'
        call rebinVector(bkgTsBinEdges, srcTsBinEdges, 'rates'&
          , real(bkgTsRate, double), real(bkgTsError, double)&
          , tempOutBkgRate, tempOutBkgError, nulls, status)
if (testFlag) print *, '**** just after rebinVector'

        outBkgRate  = tempOutBkgRate
        outBkgError = tempOutBkgError
        deallocate(tempOutBkgRate, tempOutBkgError)

        deallocate(srcTsBinEdges, bkgTsBinEdges)

        select case(status)
        case(-1)
call error('bkgTimeBinsLeZero','One or more of the bins of the background &
&time series was found to be of zero or negative width. Something is &
&seriously wrong somewhere and you probably ought to contact the code &
&developer.')
        case(-2)
call error('srcTimeBinsLeZero','One or more of the bins of the source &
&time series was found to be of zero or negative width. Something is &
&seriously wrong somewhere and you probably ought to contact the code &
&developer.')
        case(-3)
call error('bkgErrorsNegative','One or more of the background ERROR values &
&was fond to be negative.')
        case(-4)
call error('bug0',"Bug in module lccorr_pcms_mod. Please contact the code developer.")
        case default
          if (status < 0) then
call error('unknownRebinError','This ought not to happen. Please contact &
&the code developer.')
          else if (status > 0) then
            if (btest(status, 0)) then
              call warning('noOverlap', 'The background timeseries does &
              &not overlap in time at all with the source timeseries. &
              &The background columns in the output will be filled with &
              &nulls.')
            end if
            if (btest(status, 1)) then
              call warning('onlyPartialOverlap', 'The background timeseries &
              &does not entirely overlap in time with the source timeseries. &
              &The non-overlapping rows of the background columns in the &
              &output will be filled with nulls. This ought not to give &
              &problems but on the whole it is probably better to preserve &
              &the same binning scheme for the background TS as for the &
              &source TS.')
            end if
          end if
        end select

        do row = 0, numSrcTsBins - 1
          if (nulls(row+1)) then
            call setToNull(outBkgRateCol,  row)
            call setToNull(outBkgErrorCol, row)
          end if
        end do

      else ! no rebinning of the bkg ts is necessary, it lines up perfectly
      ! with the source ts.
        deallocate(bkgTsTime, bkgTsTimeDel)
        outBkgRate  = bkgTsRate
        outBkgError = bkgTsError
      end if ! rebinning of the bkg ts is necessary

      call release(bkgTsSet)
      deallocate(bkgTsRate, bkgTsError)
    end if

    call release(outSet)
    deallocate(srcTsTime, srcTsTimeDel)

  end subroutine constructOutSet

  !---------------------------------------------------------------------
  subroutine fetchTimeAndDel(tab, time, timeDel)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate pointers timeDel and time after use.

    implicit none 

    type(TableT), intent(in) :: tab
    real(single), pointer    :: timeDel(:)
    real(double), pointer    :: time(:)

    real(double) :: centreTime1stBin, timeDelScalar, empiricTimeDel
    integer(int32) :: numBins, i

    !-------------------------------------------------------------------
    ! Start:

    centreTime1stBin = calcCentreTime1stBin(tab)

    numBins = numberOfRows(tab)
    allocate(time(numBins), timeDel(numBins))
time = 0.0d0
timeDel = 0.0

    if (hasColumn(tab, 'TIME')) then
      call readColDataToFixed(tab, 'TIME', time)
      time = time + centreTime1stBin

      if (hasColumn(tab, 'TIMEDEL')) then
        call readColDataToFixed(tab, 'TIMEDEL', timeDel)

      else ! assume time bins of uniform width
        ! First use this vector to calculate the empirical time del:

!!        timeDel = cshift(time, 1) - time ! Have to throw away the
!!        ! last element since this is time(1) - time(numBins)
!timeDel(:numBins-1) = time(2:) - time(:numBins-1)

!do i = 1, numBins
!  print *, i, time(i), timeDel(i)
!end do
!print *, ''
!print *, ''
!print *, ''

!        empiricTimeDel = sum(timeDel(:numBins - 1)) / real(numBins - 1)
empiricTimeDel = (time(numBins) - time(1)) / real(numBins - 1, double)

        timeDelScalar = real64Attribute(tab, 'TIMEDEL')
        if (fracDiff(timeDelScalar, empiricTimeDel) > 0.001) then
          write(errStr, '(a, f7.5, a, f7.5, a)') 'The value of the &
            &TIMEDEL keyword, ', timeDelScalar, ', differs from the &
            &average bin width ', empiricTimeDel,' by more than 0.001.'
!call error('badTimeDel','The value of the TIMEDEL keyword is &
!&significantly different from the approximate value calculated from &
!&the time series itself.')

call error('badTimeDel')
        end if

        timeDel = timeDelScalar
      end if

    else ! get TIMEDEL keyword and use this to calculate TIMEs:
      timeDelScalar = real64Attribute(tab, 'TIMEDEL')
      timeDel = timeDelScalar
      time = centreTime1stBin + (/(real(i), i=0,numBins-1)/) * timeDelScalar
    end if

  end subroutine fetchTimeAndDel

  !---------------------------------------------------------------------
  subroutine calcTimeBinEdges(time, timeDel, binEdges)
  !---------------------------------------------------------------------
  ! NOTE! The calling routine should deallocate the pointer binEdges after use.

    implicit none 

    real(double), intent(in) :: time(:)
    real(single), intent(in) :: timeDel(size(time))
    real(double), pointer :: binEdges(:)

    integer(int32) :: numBins

    !-------------------------------------------------------------------
    ! Start:

    numBins = size(time)
    allocate(binEdges(numBins + 1))
    binEdges(1) = time(1) - timeDel(1) / 2.0
    binEdges(2:numBins) = ((time(1:numBins-1) + time(2:numBins))&
      + (timeDel(1:numBins-1) - timeDel(2:numBins)) / 2.0) / 2.0
    binEdges(numBins + 1) = time(numBins) + timeDel(numBins) / 2.0

  end subroutine calcTimeBinEdges

  !---------------------------------------------------------------------
  function calcCentreTime1stBin(tab) result(centreTime1stBin)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) :: tab
    real(double) :: centreTime1stBin

    real(double) ::&
      tsTimeZerF
    integer(int32) ::&
      tsTimeZerI

    !-------------------------------------------------------------------
    ! Start:

    if (hasAttribute(tab, 'TIMEZERI')&
    .and. hasAttribute(tab, 'TIMEZERF')) then
      tsTimeZerI = int32Attribute(tab, 'TIMEZERI')
      tsTimeZerF = real64Attribute(tab, 'TIMEZERF')
      centreTime1stBin = tsTimeZerI + tsTimeZerF ! type conversion to real64

    else if (hasAttribute(tab, 'TIMEZERO')) then
      centreTime1stBin = real64Attribute(tab, 'TIMEZERO')

    else
      centreTime1stBin = 0.0d0
    end if

  end function calcCentreTime1stBin

  !---------------------------------------------------------------------
  subroutine fetchRateAndError(inTab, timeDel, setName, rate, rateError)
  !---------------------------------------------------------------------
  ! NOTE! The calling program should deallocate the pointers rate, rateError after use.

    implicit none 

    type(TableT), intent(in) :: inTab
    real(single), intent(in) :: timeDel(:)
    character(*), intent(in) :: setName
    real(single), pointer :: rate(:), rateError(:)

    real(single), allocatable :: counts(:), countsError(:)
    integer(int32) :: numTsBins

    !-------------------------------------------------------------------
    ! Start:

    numTsBins = numberOfRows(inTab)
    allocate(rate(numTsBins), rateError(numTsBins))

    if (hasColumn(inTab, 'COUNTS')) then
      allocate(counts(numTsBins))
      call readColDataToFixed(inTab, 'COUNTS', counts)
      rate = counts / timeDel

      if (hasColumn(inTab, 'ERROR')) then
        allocate(countsError(numTsBins))
        call readColDataToFixed(inTab, 'ERROR', countsError)
        rateError = countsError / timeDel
        deallocate(countsError)

      else ! check for POISSERR keyword:
        if (.not. hasAttribute(inTab, 'POISSERR')) then
          write(errstr, *) "There is neither a POISSERR keyword nor an &
          &ERROR column in the RATE extension of dataset "//&
          trim(setName)//". However since the data are in the form of &
          &counts, it seems very likely that the uncertainties are &
          &Poissonian (ie = square root of count values). This is &
          &what lccorr is going to assume."

          call warning('tsNoPoisserrKwd')

        else if (.not. booleanAttribute(inTab, 'POISSERR')) then
          write(errstr, *) "The RATE extension header of dataset "&
          //trim(setName)//" has the POISSERR keyword set to F, &
          &but there is no ERROR column. Don't know how to deal with &
          &uncertainties."
call error('tsBadUncertaintyType')
        end if

        if (any(counts .lt. 0.0)) then
          write(errstr, *) "COUNTS value < 0 detected in the &
          &RATE extension of the timeseries "&
          //trim(setName)//". Can't calculate uncerts."
call error('tsCountsLeZero')
        end if

        rateError = sqrt(counts) / timeDel
      end if
      deallocate(counts)

    else ! doesn't have a COUNTS column, assume RATE.
      call readColDataToFixed(inTab, 'RATE', rate)

      if (hasColumn(inTab, 'ERROR')) then
        call readColDataToFixed(inTab, 'ERROR', rateError)

      else ! check for POISSERR keyword:
        if (.not. hasAttribute(inTab, 'POISSERR')) then
          write(errstr, *) "There is neither a POISSERR keyword nor an &
          &ERROR column in the RATE extension of dataset "//&
          trim(setName)//". I am going to assume that the uncertainties &
          &are Poissonian (ie = square root of RATE * bin duration)."

          call warning('tsNoPoisserrKwd')

        else if (.not. booleanAttribute(inTab, 'POISSERR')) then
          write(errstr, *) "The RATE extension header of dataset "&
          //trim(setName)//" has the POISSERR keyword set to F, &
          &but there is no ERROR column. Don't know how to deal with &
          &uncertainties."
call error('tsBadUncertaintyType')
        end if

        allocate(counts(numTsBins))
        counts = rate * timeDel
        rateError = sqrt(counts) / timeDel
        deallocate(counts)
      end if
    end if

  end subroutine fetchRateAndError

  !---------------------------------------------------------------------
  subroutine miscSetups
  !---------------------------------------------------------------------

    implicit none 

!    real(double) ::&
!      radiusFovDeg
!    integer ::&
!      patternID
!    integer(int8), pointer ::&
!      patterns(:,:,:)
    integer(int8) :: ccd, node, ccdnr
    integer(int32) ::&
      eventListInstrumentId

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, verbose, 'Some miscellaneous setups; keywords &
    &read from event list and cal state set.')

    call getCalInfo(eventSetName, eventListInstrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude) ! sets cal state for instrument.
!*** try to put all consults of the event list together because for PN there is a big time overhead.

    if (eventListInstrumentId .ne. instrumentId) then
      write(errstr, *) "INSTRUME keyword "//trim(instId(eventListInstrumentId))&
      //" in "//trim(eventSetName)//" doesn't &
      &match that of the source timeseries."
call error('evlistWrongInstrument')
    end if

    ! The next few quantities are not ccd- or node-specific:
!    call CAL_getMiscellaneousData('MM_PER_PIXEL_X', pixelWidthXmm)
!    call CAL_getMiscellaneousData('MM_PER_PIXEL_Y', pixelWidthYmm)
!    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)
!    call CAL_getMiscellaneousData('FOV_RADIUS', radiusFovDeg)

!    pixelAreaMm2 = pixelWidthXMm * pixelWidthYMm ! total area of 1 pixel
!    ! in mm^2. Implicit type conversion double -> single.

!    mmToArcsec = RadToDeg * (3600.0 / focalLengthMm)

!    mmToDetUnits = mmToArcsec / detUnitArcsec
    ! detUnitArcsec is now declared and set in ssclib/coordinate.

!    radiusFovDET = radiusFovDeg * (3600.0 / detUnitArcsec)

    if (instrumentId .eq. EMOS1&
    .or. instrumentId .eq. EMOS2) then
      chipMinRawX = EMOS_MIN_X
      chipMaxRawX = EMOS_MAX_X
      chipMinRawY = EMOS_MIN_Y
      chipMaxRawY = EMOS_MAX_Y

      numCanonicalPatterns = 13
      numCcds = 7
      numNodes = 2
!ccdMain  = 1
!nodeMain = 1
    else ! EPN.
      chipMinRawX = EPN_MIN_X
      chipMaxRawX = EPN_MAX_X
      chipMinRawY = EPN_MIN_Y
      chipMaxRawY = EPN_MAX_Y

      numCanonicalPatterns = 5
      numCcds = 12
      numNodes = 1
!ccdMain  = 4
!nodeMain = 1
    end if
    ! Defined in caloalutils.f90.

    !....................................................................
    ! Load some handy instrument-specific arrays:

    do ccd = 1, numCcds
      do node = 1, numNodes
        ccdnr = 10 * (node - 1) + ccd
        ccd2ccdnr(ccd, node) = ccdnr
        ccdnr2ccd(ccdnr) = ccd
        ccdnr2node(ccdnr) = node

        write(ccdStr(ccd, node), '(i2.2)') ccdnr
      end do
    end do

  end subroutine miscSetups

  !---------------------------------------------------------------------
  subroutine checkTsDss(tsSetName)
  !---------------------------------------------------------------------

    implicit none

    character(*), intent(in) :: tsSetName

    type(DataSetT) :: set
    type(BlockT) :: tsBlock
    type(DssT) :: dssPointer
    integer :: numComps, compNum, numFilts, filtNum, i, filtType
    type(DScompT) :: dssComp
    type(DSfilterT) :: dssFilt
    character(80) :: columnNames, xColName, yColName
    type(BitMaskT), pointer :: flagFilter(:) => null()
    integer(int32) :: onBits, offBits, offBitsWarningFlag, onnBitsWarningFlag
    integer(int8) :: ccdNum, nodeNum, ccdNodeNum
    integer, pointer :: dssCompNums(:) => null()
    logical(bool) :: noCcdsPassDss

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, ' aaa checkTsDss'

    call message(AppMsg, Verbose, "Checking the data subspace of "&
      //trim(tsSetName))

    set = dataSet(tsSetName, READ)
    tsBlock = block(set, 'RATE')
if (.not. hasDss(tsBlock)) call error('noDss',"The dataset "//trim(tsSetName)&
  //" has no data subspace. Lccorr cannot determine which filtering &
  &criteria were used without this.")
if (testFlag) print *, ' bbb'

    dssPointer = dataSubspace(tsBlock)
if (testFlag) print *, ' ccc'
    numComps = dssComponents(dssPointer)
if (numComps <= 0) call error('noDssComponents',"The data subspace of &
  &the dataset "//trim(tsSetName)//" has 0 components. Lccorr cannot &
  &determine which filtering criteria were used in this case.")
if (testFlag) print *, ' ddd'

    offBitsWarningFlag = 0
    onnBitsWarningFlag = 0
if (testFlag) print *, ' eee'
    do compNum = 0, numComps - 1
if (testFlag) print *, 'compNum', compNum
      dssComp = dssComponent(dssPointer, compNum)
      numFilts = dssFilters(dssComp)
      do filtNum = 0, numFilts - 1
if (testFlag) print *, '  filtNum', filtNum
        dssFilt = dssFilter(dssComp, filtNum)

        ! Check for 2d filters on any but xy pairs:

        filtType = dssFilterType(dssFilt)
        if (filtType == REGIONFILTERT .or. filtType == MASKFILTERT) then
          columnNames = dssFilterComps(dssFilt)
          call decomposeRegionXyNameString(columnNames, xColName, yColName)
if (testFlag) print *, '    column names:', trim(xColName)//' '//trim(yColName)

          if (.not. ((xColName .eq.    'X' .and. yColName .eq.    'Y')&
                .or. (xColName .eq. 'RAWX' .and. yColName .eq. 'RAWY')&
                .or. (xColName .eq. 'DETX' .and. yColName .eq. 'DETY'))) then
            call warning('illegalTwoAxisDssFilter',"The data subspace &
              &contains a two-axis filter on "//trim(xColName)//" and "&
              //trim(yColName)//". Lccorr can only handle filters on &
              &X/Y, RAWX/RAWY or DETX/DETY.")
          end if

        else ! single-axis filter.
          ! Check for filters on PHA:

          columnNames = dssFilterName(dssFilt)
if (testFlag) print *, '    column name:', trim(columnNames)
          if (index(columnNames, 'PHA') > 0)&
            call warning('illegalDssFilterAxis',"The data subspace contains &
              &a filter on column PHA of the event list: lccorr cannot &
              &handle this.")

          if (index(columnNames, 'FLAG') > 0) then
            ! Check for REJECTED_BY_GATTI, OUTSIDE_THRESHOLDS, BAD_E3E4,
            ! UNDERSHOOT, IN_BAD_FRAME and IN_SPOILED_FRAME flag selections:

            call getBitMasks(dssFilt, flagFilter) ! in ssclib/dss_bitmasks_aux.

            do i = 1, size(flagFilter)
              onBits  = flagFilter(i)%onBits
              offBits = flagFilter(i)%offBits

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_DIAGONAL))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_IN_SPOILED_FRAME))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_UNKNOWN_RAWX_PATTERN))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_IN_BAD_FRAME))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_COSMIC_RAY))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_REJECTED_BY_GATTI))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_OUTSIDE_THRESHOLDS))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_UNDERSHOOT))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_BAD_E3E4))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_IN_SPOILED_FRAME))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_INVALID_PATTERN))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_SECONDARY))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_TRAILING))

              offBitsWarningFlag = ior(offBitsWarningFlag, iand(offBits&
                , EVATT_MIP_ASSOCIATED))


              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_DIAGONAL))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_IN_SPOILED_FRAME))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_UNKNOWN_RAWX_PATTERN))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_IN_BAD_FRAME))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_COSMIC_RAY))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_REJECTED_BY_GATTI))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_OUTSIDE_THRESHOLDS))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_UNDERSHOOT))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_BAD_E3E4))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_IN_SPOILED_FRAME))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_INVALID_PATTERN))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_SECONDARY))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_TRAILING))

              onnBitsWarningFlag = ior(onnBitsWarningFlag, iand(onBits&
                ,  EVATT_MIP_ASSOCIATED))
            end do ! loop over elements of flagFilter
            deallocate(flagFilter)
          end if ! filter on 'FLAG'
        end if ! single-axis filter.
      end do ! filtNum
    end do ! compNum
if (testFlag) print *, ' fff'

offBitsWarningFlag = 0
onnBitsWarningFlag = 0
!*** just temporarily until I get my head around this section. I am not sure I ahve done it right, especially the onnBits.

if (testFlag) print *, ' ggg'
    if (offBitsWarningFlag > 0) then
      call warning('uncorrectedEventAttribute',"Some events with the &
        &following FLAG settings were excluded when the time series "&
        //trim(tsSetName)//" was made:")

      if  (iand(offBitsWarningFlag, EVATT_DIAGONAL) > 0)&
        call warning('uncorrectedEventAttribute',"  DIAGONAL")

      if  (iand(offBitsWarningFlag, EVATT_IN_SPOILED_FRAME) > 0)&
        call warning('uncorrectedEventAttribute',"  IN_SPOILED_FRAME")

      if  (iand(offBitsWarningFlag, EVATT_UNKNOWN_RAWX_PATTERN) > 0)&
        call warning('uncorrectedEventAttribute',"  UNKNOWN_RAWX_PATTERN")

      if  (iand(offBitsWarningFlag, EVATT_IN_BAD_FRAME) > 0)&
        call warning('uncorrectedEventAttribute',"  IN_BAD_FRAME")

      if  (iand(offBitsWarningFlag, EVATT_COSMIC_RAY) > 0)&
        call warning('uncorrectedEventAttribute',"  COSMIC_RAY")

      if  (iand(offBitsWarningFlag, EVATT_REJECTED_BY_GATTI) > 0)&
        call warning('uncorrectedEventAttribute',"  REJECTED_BY_GATTI")

      if  (iand(offBitsWarningFlag, EVATT_OUTSIDE_THRESHOLDS) > 0)&
        call warning('uncorrectedEventAttribute',"  OUTSIDE_THRESHOLDS")

      if  (iand(offBitsWarningFlag, EVATT_UNDERSHOOT) > 0)&
        call warning('uncorrectedEventAttribute',"  UNDERSHOOT")

      if  (iand(offBitsWarningFlag, EVATT_BAD_E3E4) > 0)&
        call warning('uncorrectedEventAttribute',"  BAD_E3E4")

      if  (iand(offBitsWarningFlag, EVATT_IN_SPOILED_FRAME) > 0)&
        call warning('uncorrectedEventAttribute',"  IN_SPOILED_FRAME")

      if  (iand(offBitsWarningFlag, EVATT_INVALID_PATTERN) > 0)&
        call warning('uncorrectedEventAttribute',"  INVALID_PATTERN")

      if  (iand(offBitsWarningFlag, EVATT_SECONDARY) > 0)&
        call warning('uncorrectedEventAttribute',"  SECONDARY")

      if  (iand(offBitsWarningFlag, EVATT_TRAILING) > 0)&
        call warning('uncorrectedEventAttribute',"  TRAILING")

      if  (iand(offBitsWarningFlag, EVATT_MIP_ASSOCIATED) > 0)&
        call warning('uncorrectedEventAttribute',"  MIP_ASSOCIATED")

      call warning('uncorrectedEventAttribute',"The task cannot yet &
        &correct for exposure lost due to the exclusion of events with &
        &these FLAG values. However, none of these is likely to &
        &produce significantly wrong results.")
    end if
if (testFlag) print *, ' hhh'

    if (onnBitsWarningFlag > 0) then
      call warning('deprecatedEventSelection',"Some events with the &
        &following FLAG settings were included when the time series "&
        //trim(tsSetName)//" was made:")

      if (iand(onnBitsWarningFlag,  EVATT_DIAGONAL) > 0)&
        call warning('deprecatedEventSelection',"  DIAGONAL")

      if (iand(onnBitsWarningFlag,  EVATT_IN_SPOILED_FRAME) > 0)&
        call warning('deprecatedEventSelection',"  IN_SPOILED_FRAME")

      if (iand(onnBitsWarningFlag,  EVATT_UNKNOWN_RAWX_PATTERN) > 0)&
        call warning('deprecatedEventSelection',"  UNKNOWN_RAWX_PATTERN")

      if (iand(onnBitsWarningFlag,  EVATT_IN_BAD_FRAME) > 0)&
        call warning('deprecatedEventSelection',"  IN_BAD_FRAME")

      if (iand(onnBitsWarningFlag,  EVATT_COSMIC_RAY) > 0)&
        call warning('deprecatedEventSelection',"  COSMIC_RAY")

      if  (iand(onnBitsWarningFlag,  EVATT_REJECTED_BY_GATTI) > 0)&
        call warning('deprecatedEventSelection',"  REJECTED_BY_GATTI")

      if  (iand(onnBitsWarningFlag,  EVATT_OUTSIDE_THRESHOLDS) > 0)&
        call warning('deprecatedEventSelection',"  OUTSIDE_THRESHOLDS")

      if  (iand(onnBitsWarningFlag,  EVATT_UNDERSHOOT) > 0)&
        call warning('deprecatedEventSelection',"  UNDERSHOOT")

      if  (iand(onnBitsWarningFlag,  EVATT_BAD_E3E4) > 0)&
        call warning('deprecatedEventSelection',"  BAD_E3E4")

      if  (iand(onnBitsWarningFlag,  EVATT_IN_SPOILED_FRAME) > 0)&
        call warning('deprecatedEventSelection',"  IN_SPOILED_FRAME")

      if  (iand(onnBitsWarningFlag,  EVATT_INVALID_PATTERN) > 0)&
        call warning('deprecatedEventSelection',"  INVALID_PATTERN")

      if  (iand(onnBitsWarningFlag,  EVATT_SECONDARY) > 0)&
        call warning('deprecatedEventSelection',"  SECONDARY")

      if  (iand(onnBitsWarningFlag,  EVATT_TRAILING) > 0)&
        call warning('deprecatedEventSelection',"  TRAILING")

      if  (iand(onnBitsWarningFlag,  EVATT_MIP_ASSOCIATED) > 0)&
        call warning('deprecatedEventSelection',"  MIP_ASSOCIATED")

      call warning('deprecatedEventSelection',"It is not recommended to &
        &include such events in a time series (light curve) of cosmic x-rays.")
    end if

    ! Foreach ccd/node, check that all filters on TIME are the same for the
    ! whole ccd.
if (testFlag) print *, ' iii'

    noCcdsPassDss = .true. ! default
    do ccdNum = 1, numCcds     ! numCcds, numNodes and ccd2ccdnr are declared
if (testFlag) print *, 'ccdNum:', ccdNum
      do nodeNum = 1, numNodes ! in lccorr_include, and their values are set
                               ! in miscSetups.
if (testFlag) print *, '  nodeNum:', nodeNum

        ccdNodeNum = ccd2ccdnr(ccdNum, nodeNum)
        call dssCompNumsWhichPassThisValue(dssPointer, 'CCDNR'&
          , real(ccdNodeNum, single), dssCompNums)

      if (size(dssCompNums) <= 0) cycle

        noCcdsPassDss = .false.

if (testFlag) print *, '    size(dssCompNums):', size(dssCompNums)
if (testFlag) print *, '    dssCompNums:', dssCompNums

        ! Check that filters on TIME are all the same for these dssCompNums.
        ! (Should also check any time-related filters on FLAG, but it could
        ! get a bit complicated, so we won't do this yet.)

        if (.not. dssCompsIdentical(dssPointer, dssCompNums, 'CCDNR'))&
          call warning('timeSelectionTooComplicated',"Lccorr can handle &
            &only 1 type of time filter per CCD. However, a more &
            &complicated time-filtering scheme was found in the data &
            &subspace.")

        deallocate(dssCompNums)
if (testFlag) print *, ''
      end do
    end do
if (testFlag) print *, ' jjj'

    call release(set)
    call dssRelease(dssPointer)

if (noCcdsPassDss) call error('noCcdsPassDss',"The data subspace of the &
  &dataset will not pass any value of CCDNR.")
if (testFlag) print *, ' zzz checkTsDss'

  end subroutine checkTsDss

  !---------------------------------------------------------------------
  subroutine getSrcPositions(wcs, srcRaDeg, srcDecDeg)
  !---------------------------------------------------------------------

    implicit none 

    type(WcsT),   intent(in)  :: wcs
    real(double), intent(out) ::&
      srcRaDeg,&
      srcDecDeg

    type(DataSetT) :: set
    logical :: succeeded
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, '  aaa getSrcPositions'

    select case (srcPositionStyle)
    case ('user')
      srcRaDeg  = userSrcRaDeg
      srcDecDeg = userSrcDecDeg

    case ('keyword')
      set = dataSet(srcPositionSetName, READ)
      srcRaDeg  = real64Attribute(set, 'SRC_RA')
      srcDecDeg = real64Attribute(set, 'SRC_DEC')
      call release(set)

    case ('dss')
      call attemptToGetSrcPosFromDss(wcs, succeeded, srcRaDeg, srcDecDeg)
if (.not. succeeded) call error('unableGetSrcPosFromDss',"You requested &
  &the task to estimate the position of the source from the data subspace &
  &(--srcposstyle=dss). However the data subspace was in too complicated &
  &form for the task to manage this. You will have to put in the source &
  &position directly (set --srcposstyle=user and supply --srcra and &
  &--srcdec).")

    case default
call error('badSrcPosStyle',"The value of --srcposstyle was not recognized.")
    end select

if (testFlag) print *, '  zzz getSrcPositions'

  end subroutine getSrcPositions

  !---------------------------------------------------------------------
  subroutine attemptToGetSrcPosFromDss(wcs, succeeded, srcRaDeg, srcDecDeg)
  !---------------------------------------------------------------------
  ! This is a fairly dumb routine. It looks through the DSS and quits at
  ! the first circle it finds. The centre of this circle is returned as
  ! the source position.

    implicit none 

    type(WcsT),   intent(in)  :: wcs
    logical,      intent(out) :: succeeded
    real(double), intent(out) :: srcRaDeg,&
                                 srcDecDeg

    type(DssT)      :: dssPointer
    type(DScompT)   :: dssComp
    integer         :: numComps, compNum, numRegions, row
    type(DataSetT) :: set
    type(BlockT) :: rateBlock
    type(DSfilterT) :: filter
    character(255) :: blockName
    type(TableT) :: regionTab
    type(ColumnT) :: shapeCol, xCol, yCol
    real(single), pointer :: x(:), y(:)
    real(single) :: srcRawX, srcRawY
    real(double) :: srcDetX, srcDetY
!    integer(int8) :: ccdNum
integer(int8) :: ccdNodeNum, ccdNum, nodeNum
    real(double) :: tanX, tanY
    real(double) :: srcRa, srcDec
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, 'aaa attemptToGetSrcPosFromDss'
    call message(AppMsg, Verbose, 'Attempting to estimate the position &
      &of the source from the data subspace.')

    succeeded = .false. ! default
    srcRaDeg  = 0.0d0
    srcDecDeg = 0.0d0

    set = dataSet(srcTsSetName, READ)
    rateBlock = block(set, 'RATE')
  if (.not. hasDss(rateBlock)) return ! hasDss is in ssclib/dss_aux.

    dssPointer = dataSubspace(rateBlock)
    numComps = dssComponents(dssPointer)
    if (numComps <= 0) then
      call dssRelease(dssPointer)
  return
    end if

if (testFlag) print *, 'bbb'
    COMPLOOP: do compNum = 0, numComps - 1
if (testFlag) print *, '  ccc compNum:', compNum
      dssComp = dssComponent(dssPointer, compNum)

      call getCcdNumFromDssComp(dssComp, numCcds, numNodes, ccdNodeNum)
      ccdNum = ccdnr2ccd(ccdNodeNum)
      nodeNum = ccdnr2node(ccdNodeNum)
!      call CAL_setState(ccdChipId=int(ccdNum, int32))
      if (instrumentId == EPN) then ! just to stop the cal giving a warning:
        call CAL_setState(ccdChipId=int(ccdNum, int32))
      else
        call CAL_setState(ccdChipId=int(ccdNum, int32)&
          , ccdNodeId=int(nodeNum - 1, int32))
      end if

      if (dssHasFilter(dssComp, 'RAW', REGIONBLOCKFILTERT)) then
if (testFlag) print *, '  ddd'
        filter = dssFilter(dssComp, 'RAW', REGIONBLOCKFILTERT)
        blockName = dssFilterBlock(filter)
        regionTab = table(set, blockName) ! assume the TS is the region dataset.
        numRegions = numberOfRows(regionTab)
        shapeCol = column(regionTab, 'SHAPE')
        do row = 0, numRegions - 1
if (testFlag) print *, '    eee row:', row
          if (stringCell(shapeCol, row) == 'CIRCLE') then
            xCol = column(regionTab, 'RAWX')
            yCol = column(regionTab, 'RAWY')
            x => real32Data(xCol)
            y => real32Data(yCol)
!*** should use readColData, it would be simpler.
            srcRawX = x(row)
            srcRawY = y(row)
            call release(regionTab)

if (testFlag) print *, '    eee0'
            call rawTodet(nint(srcRawX, int16), nint(srcRawY, int16)&
              , srcDetX, srcDetY)
if (testFlag) print *, '    eee1'

            call detToRaDec(srcDetX, srcDetY&
              , scAttitude, expStartTimeStamp, srcRa, srcDec)
if (testFlag) print *, '    eee2'
            srcRaDeg  = srcRa  / PiOn180
            srcDecDeg = srcDec / PiOn180

            succeeded = .true.
    exit COMPLOOP
          end if
        end do
        call release(regionTab)
      end if
if (testFlag) print *, '  fff'

      if (dssHasFilter(dssComp, 'DET', REGIONBLOCKFILTERT)) then
if (testFlag) print *, '  ggg'
        filter = dssFilter(dssComp, 'DET', REGIONBLOCKFILTERT)
        blockName = dssFilterBlock(filter)
        regionTab = table(set, blockName) ! assume the TS is the region dataset.
        numRegions = numberOfRows(regionTab)
        shapeCol = column(regionTab, 'SHAPE')
        do row = 0, numRegions - 1
if (testFlag) print *, '    hhh row:', row
          if (stringCell(shapeCol, row) == 'CIRCLE') then
            xCol = column(regionTab, 'DETX')
            yCol = column(regionTab, 'DETY')
            x => real32Data(xCol)
            y => real32Data(yCol)
!*** should use readColData, it would be simpler.
            srcDetX = real(x(row), double)
            srcDetY = real(y(row), double)
            call release(regionTab)

if (testFlag) print *, '    hhh0'
            call detToRaDec(srcDetX, srcDetY&
              , scAttitude, expStartTimeStamp, srcRa, srcDec)
if (testFlag) print *, '    hhh1'
            srcRaDeg  = srcRa  / PiOn180
            srcDecDeg = srcDec / PiOn180

            succeeded = .true.
    exit COMPLOOP
          end if
        end do
        call release(regionTab)
      end if
if (testFlag) print *, '  iii'

      if (dssHasFilter(dssComp, 'POS', REGIONBLOCKFILTERT)) then
if (testFlag) print *, '  jjj'
        filter = dssFilter(dssComp, 'POS', REGIONBLOCKFILTERT)
        blockName = dssFilterBlock(filter)
        regionTab = table(set, blockName) ! assume the TS is the region dataset.
        numRegions = numberOfRows(regionTab)
        shapeCol = column(regionTab, 'SHAPE')
        do row = 0, numRegions - 1
          if (stringCell(shapeCol, row) == 'CIRCLE') then
            xCol = column(regionTab, 'X')
            yCol = column(regionTab, 'Y')
            x => real32Data(xCol)
            y => real32Data(yCol)
!*** should use readColData, it would be simpler.

if (testFlag) print *, '    jjj0'
if (testFlag) print *, 'wcs%x%worldUnit, wcs%y%worldUnit:', wcs%x%worldUnit, wcs%y%worldUnit
if (testFlag) print *, 'wcs%x%ctype, wcs%y%ctype:', wcs%x%ctype, wcs%y%ctype

            call pixelsToTan(x(row), y(row), wcs, tanX, tanY)
            ! in ssclib/wcs_aux
if (testFlag) print *, '    jjj1'

            call tanToRaDec(tanX, tanY, wcs%x%refWorld*PiOn180&
              , wcs%y%refWorld*PiOn180, srcRa, srcDec) ! in ssclib/coordinate
if (testFlag) print *, '    jjj2'

!*** assumes RA/DEC TAN type, and also that worldUnits are both 'deg'. This should be checked.
srcRaDeg  = srcRa  / PiOn180
srcDecDeg = srcDec / PiOn180

            succeeded = .true.
    exit COMPLOOP
          end if
        end do
        call release(regionTab)
      end if
    end do COMPLOOP
if (testFlag) print *, 'zzz attemptToGetSrcPosFromDss'

    call dssRelease(dssPointer)

  end subroutine attemptToGetSrcPosFromDss

  !---------------------------------------------------------------------
  subroutine getCcdNumFromDssComp(dssComp, numCcds, numNodes, ccdNodeNum)
  !---------------------------------------------------------------------

    implicit none 

    type(DScompT), intent(in) :: dssComp
    integer(int8), intent(in) :: numCcds, numNodes
    integer(int8), intent(out) :: ccdNodeNum

    type(DSfilterT) :: firstFilter
    integer(int8) :: ccdNum, nodeNum, testCcdNodeNum
    logical :: noCcdFilterFound

    !-------------------------------------------------------------------
    ! Start:

    if (numFiltersOfAxis(dssComp, 'CCDNR') <= 0)&
call error('noCcdFiltersInDssComponent',"There is no filter on CCDNR in &
  &the data subspace.")

    call get1stFilterThisAxis(dssComp, 'CCDNR', firstFilter)

    noCcdFilterFound = .true. ! default
    do ccdNum = 1, numCcds
      do nodeNum = 1, numNodes
!        testCcdNodeNum = ccd2ccdnr(ccd, node)
testCcdNodeNum = 10 * (nodeNum - 1) + ccdNum

        if(valueIsPassedByFilter(real(testCcdNodeNum,single),firstFilter))then
          ccdNodeNum = testCcdNodeNum
          noCcdFilterFound = .false.
  return
        end if
      end do
    end do

    if (noCcdFilterFound) then
      write(errstr, '(a, i2, a)') "There is no DSS filter on ccd/node &
        &number within the specified range."
call error('noFilterThisCcdInDssComponent')
    end if

  end subroutine getCcdNumFromDssComp

  !---------------------------------------------------------------------
  subroutine calcWeightListStr(ccdWeights, weightListStr)
  !---------------------------------------------------------------------

    implicit none 

    real(single), intent(in)  :: ccdWeights(:)
    character(*), intent(out) :: weightListStr

    integer :: localNumCcds, ccdNum

    !-------------------------------------------------------------------
    ! Start:

    localNumCcds = size(ccdWeights)
if (len(weightListStr) < 7*localNumCcds-1) then
print *, len(weightListStr), localNumCcds, 7*localNumCcds-1
call error('stringTooShort',"")
end if

    weightListStr = ''
    do ccdNum = 1, localNumCcds
      if (ccdNum == 1) then
        write(weightListStr, '(f6.4)') ccdWeights(ccdNum)
      else
        write(weightListStr, '(a, a1, f6.4)') trim(weightListStr), ' '&
          , ccdWeights(ccdNum)
      end if
    end do

  end subroutine calcWeightListStr

  !---------------------------------------------------------------------
  subroutine calcWeightListStr_20040923(ccdWeights, weightListStr)
  !---------------------------------------------------------------------

    implicit none 

    real(single), intent(in)  :: ccdWeights(:,:)
    character(*), intent(out) :: weightListStr

    integer ::&
      localNumCcds, localNumNodes, localMaxCcdNodeNum,&
      ccdNum, nodeNum, ccdNodeNum

    !-------------------------------------------------------------------
    ! Start:

    localNumCcds = size(ccdWeights,1)
    localNumNodes = size(ccdWeights,2)
    localMaxCcdNodeNum = 10 * (localNumNodes - 1) + localNumCcds
if (len(weightListStr) < 7*localMaxCcdNodeNum-1) then
print *, len(weightListStr), localNumNodes, localNumCcds, localMaxCcdNodeNum, 7*localMaxCcdNodeNum-1
call error('stringTooShort',"")
end if

    weightListStr = ''
    do ccdNum = 1, localNumCcds
      do nodeNum = 1, localNumNodes
        ccdNodeNum = 10 * (nodeNum - 1) + ccdNum

        if (ccdNodeNum == 1) then
          write(weightListStr, '(f6.4)') ccdWeights(ccdNum,nodeNum)
        else
          write(weightListStr, '(a, a1, f6.4)') trim(weightListStr), ' '&
            , ccdWeights(ccdNum,nodeNum)
        end if
      end do
    end do

  end subroutine calcWeightListStr_20040923

  !---------------------------------------------------------------------
  subroutine rescaleBkg(outTab, ratioBkgOverSrcAreas)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: outTab
    real(single), intent(in) :: ratioBkgOverSrcAreas
      
    type(ColumnT) :: outBkgRateCol, outBkgErrorCol
    real(single), pointer :: outBkgRate(:), outBkgError(:)

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Normalising bkg to src collection area.')

    outBkgRateCol  = column(outTab, 'BRATE',  MODIFY)
    outBkgErrorCol = column(outTab, 'BERROR', MODIFY)

    outBkgRate  => real32data(outBkgRateCol)
    outBkgError => real32data(outBkgErrorCol)

    ! Vector ops:
    outBkgRate  = outBkgRate  / ratioBkgOverSrcAreas
    outBkgError = outBkgError / ratioBkgOverSrcAreas

    call release(outBkgRateCol)
    call release(outBkgErrorCol)

  end subroutine rescaleBkg

  !---------------------------------------------------------------------
  subroutine subtractCompensatedBkg(outTab)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in) :: outTab
      
    integer(int32) :: row, numTsBins
    type(ColumnT) ::&
      outRateCol, outErrorCol, outBkgRateCol, outBkgErrorCol
    real(single), pointer ::&
      outRate(:), outError(:), outBkgRate(:), outBkgError(:)

    !-------------------------------------------------------------------
    ! Start:

    numTsBins = numberOfRows(outTab)

    outBkgRateCol  = column(outTab, 'BRATE',  MODIFY)
    outBkgErrorCol = column(outTab, 'BERROR', MODIFY)
    outBkgRate  => real32data(outBkgRateCol)
    outBkgError => real32data(outBkgErrorCol)

    !....................................................................
    ! Subtract background from source:

    call message(AppMsg, Verbose, 'Subtracting bkg from src TS.')

    outRateCol  = column(outTab, 'RATE', MODIFY)
    outErrorCol = column(outTab, 'ERROR', MODIFY)
    outRate  => real32data(outRateCol)
    outError => real32data(outErrorCol)
    do row = 0, numTsBins - 1
      if (.not. isNull(outRateCol, row)) then
        if (isNull(outBkgRateCol, row)) then
          call setToNull(outRateCol, row)
          call setToNull(outErrorCol, row)
        else
          outRate(row) = outRate(row) - outBkgRate(row)
          outError(row) = sqrt&
            ( outError(row)    * outError(row)&
            + outBkgError(row) * outBkgError(row))
        end if
      end if
    end do

    call release(outBkgRateCol)
    call release(outBkgErrorCol)
    call release(outRateCol)
    call release(outErrorCol)

  end subroutine subtractCompensatedBkg

  !---------------------------------------------------------------------
  subroutine finaliseOutSet
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) ::&
      outSet
    type(TableT)   ::&
      outTab
    type(ColumnT)  ::&
      outBkgRateCol,&
      outBkgErrorCol
integer :: ioUnit, ioStatus

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

if (testFlag) print *, 'aaaa finaliseOutSet'
    call message(AppMsg, Verbose, 'Cleaning up output TS.')

    outSet = clone(tempOutSetName, outSetName, MODIFY)
!outSet = dataSet(tempOutSetName, MODIFY)
!call rename(outSet, outSetName)

if (testFlag) print *, 'eeee'
    outTab = table(outSet, 'RATE')
if (testFlag) print *, 'ffff'

    if (withBkgTsSet) then
      outBkgRateCol = column(outTab, 'BRATE', MODIFY)
      call rename(outBkgRateCol, 'BACKV')
      call release(outBkgRateCol)
      outBkgErrorCol = column(outTab, 'BERROR', MODIFY)
      call rename(outBkgErrorCol, 'BACKE')
      call release(outBkgErrorCol)
    end if

    if (hasColumn(outTab, 'BDEADC')) then
      call deleteColumn(outTab, 'BDEADC')
    end if
    if (hasColumn(outTab, 'BVIGNET')) then
      call deleteColumn(outTab, 'BVIGNET')
    end if
    if (hasColumn(outTab, 'BFRACEXP')) then
      call deleteColumn(outTab, 'BFRACEXP')
    end if

if (testFlag) print *, 'gggg'
    call release(outSet)

call getFreeIoUnit(ioUnit) ! in ssclib/ssc_misc.
open(ioUnit, iostat=ioStatus, file=tempOutSetName, status='old')
if (ioStatus .ne. 0) call error('openFileError',"")
close(ioUnit, iostat=ioStatus, status='delete')
if (ioStatus .ne. 0) call error('closeFileError',"")


if (testFlag) print *, 'zzzz finaliseOutSet'

  end subroutine finaliseOutSet

  !---------------------------------------------------------------------
  subroutine memoryLeakTest
  !---------------------------------------------------------------------

    implicit none 

    real(double), pointer :: fred(:,:)!, ernie(:,:)

    !-------------------------------------------------------------------
    ! Start:

    allocate(fred(1000,1000))
    do
      fred => modifyFred(fred)
    end do

  end subroutine memoryLeakTest

  !---------------------------------------------------------------------
  function modifyFred(inPtr) result(outPtr)
  !---------------------------------------------------------------------

    implicit none 

    real(double), pointer :: inPtr(:,:), outPtr(:,:)

    !-------------------------------------------------------------------
    ! Start:

    allocate(outPtr(1000,1000))
    deallocate(inPtr)

  end function modifyFred

  !---------------------------------------------------------------------
  subroutine dumpModeParams(modeParams)
  !---------------------------------------------------------------------

    implicit none 

    type(ModeParameterDataType), intent(in) :: modeParams

    !-------------------------------------------------------------------
    ! Start:

    print *, ''
    print *, '        Mode param values:'
    print *, '-----------------------------------'
    print *, '  integrationTime:', modeParams%integrationTime
    print *, '        shiftTime:', modeParams%shiftTime
    print *, '        frameTime:', modeParams%frameTime
    print *, '        cycleTime:', modeParams%cycleTime
    print *, '   timeResolution:', modeParams%timeResolution
    print *, '        dutyCycle:', modeParams%dutyCycle
    print *, '  smearedFraction:', modeParams%smearedFraction
    print *, '-----------------------------------'

  end subroutine dumpModeParams

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module lccorr_pcms_mod
